/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/cli/invocation/CliInvocation.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.cli.invocation;

import java.lang.reflect.InvocationTargetException;

import org.apache.commons.cli2.OptionException;

import com.netapp.common.MsgKey;
import com.netapp.common.cli.BindingException;
import com.netapp.common.cli.CliParser;
import com.netapp.common.cli.Command;
import com.netapp.common.cli.binding.CommandBinding;
import com.netapp.common.cli.validator.ValidationException;
import com.netapp.common.logging.NALogger;

/**
 * Runs a CLI command (arg array) against a {@link CliInvocable}.
 * Typically called from main() methods
 * 
 * <pre>
 *  public static void main(String[] args) {
 *       CLIFacadeInvocation<SmoServerFacade> cliFacadeInvocation = new CLIFacadeInvocation<SmoServerFacade>(args, new SmoServerFacade(), logger);
 *       
 *       int exitCode = cliFacadeInvocation.invoke();
 *       if (exitCode != 0) {
 *           System.exit(exitCode);
 *       }
 *   }
 * </pre>
 * 
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class CliInvocation {

    private static final NALogger logger = NALogger.getLogger();
    
    private String[] args;
    private CliInvocable invocable;
    
    public CliInvocation(String[] args, CliInvocable invocable) {
        this.args = args;
        this.invocable = invocable;
    }
    
    public int invoke() {

        try {
            CliParser cliParser = new CliParser(this.invocable.getClass());
            
            CommandBinding commandBinding = cliParser.parse(this.args);
            commandBinding.setGlobalFields(this.invocable);
            commandBinding.execute(this.invocable);
            return this.invocable.exitCode();
        }
        catch (InvocationTargetException e) {
            logger.error(MsgKey.CLI_INVOCATION_INVOCATION_ERROR, e, this.niceThrowableMessage(e.getCause()));
            return 1;
        }
        catch (OptionException e) {
            logger.error(MsgKey.CLI_INVOCATION_PARSE_ERROR, e, e.getLocalizedMessage());
            showUsage(this.invocable, this.args);
            return 1;
        }
        catch (BindingException e) {
            logger.error(MsgKey.CLI_INVOCATION_PARSE_ERROR, e, e.getLocalizedMessage());
            showUsage(this.invocable, this.args);
            return 1;
        }
        catch (ValidationException e) {
            logger.error(MsgKey.CLI_INVOCATION_VALIDATION_ERROR, e, e.getLocalizedMessage());
            showUsage(this.invocable, this.args);
            return 1;
        }
        /*
         * CHECKSTYLE:OFF
         * We need to catch all throwables here so as not to let them up into root of the JVM
         */
        catch (Throwable t) {
            // CHECKSTYLE:ON
            logger.error(MsgKey.CLI_INVOCATION_UNEXPECTED_ERROR, t, this.niceThrowableMessage(t));
            showUsage(this.invocable, this.args);
            return 1;
        }
    }
    
    private void showUsage(CliInvocable cliInvocable, String[] args) {
        try {
            if (args.length > 0) {
                cliInvocable.help(new Command(args[0]));    
            }
            else {
                cliInvocable.help();
            }
        }
        catch (IllegalArgumentException e) {
            // The user typed an invalid top-level SMO command
            cliInvocable.help();
        }
    }
 
    private String niceThrowableMessage(Throwable t) {
        return t.getLocalizedMessage() != null ? t.getLocalizedMessage() : t.toString();
    }
}

