/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/credentials/service/LocalCredentialService.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc. 
 * All rights reserved.
 */
package com.netapp.common.credentials.service;

import java.security.PrivateKey;
import java.security.PublicKey;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Hex;

import com.netapp.common.MsgKey;
import com.netapp.common.credentials.Credential;
import com.netapp.common.credentials.CredentialKey;
import com.netapp.common.credentials.CredentialPersistenceException;
import com.netapp.common.credentials.CredentialPersistenceManager;
import com.netapp.common.encryption.Encryption;
import com.netapp.common.encryption.EncryptionException;
import com.netapp.common.flow.api.OperationResult;
import com.netapp.common.flow.api.OperationState;
import com.netapp.common.logging.LogLevel;
import com.netapp.common.logging.LogMessage;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.StringUtils;

/**
 * Implement CredentialService using the frameworks CredentialPersistenceManager.
 * 
 * The following configuration properties are expected to be defined (or use default value):
 * 
 * <ul>
 *   <li>credential.asymmetric.private.path - Path to private key file for asymmetric key
 *   <li>credential.asymmetric.algorithm - Asymmetric algorithm to use. Defaults to RSA
 *   <li>credential.service.version.release - Build release number
 *   <li>credential.service.version.interface - Version number for the SOAP interface
 * </ul>
 * 
 * @author joshb
 *
 */
public class LocalCredentialService implements CredentialService {
    private static final NALogger LOGGER = NALogger.getLogger(LocalCredentialService.class);

    private ConfigBundle context;
    private PrivateKey serverPrivateKey;
    private PublicKey clientPublicKey;
    private String serverAlgorithm;
    private String clientAlgorithm;
    private CredentialPersistenceManager credentialManager;

    public LocalCredentialService() {
    }

    /** 
     * Init method. To be called after all configuration is complete, generally
     * by Spring or another framework.
     *
     */
    public void init() {
        if (null != getContext()) {

            String serverPrivateKeyPath = CredentialServiceConstants.SERVER_PRIVATE_KEY_PATH.lookup(getContext());
            String serverSideAlgorithm = CredentialServiceConstants.SERVER_ENCRYPTION_ALGORITHM.lookup(getContext());
            String clientPublicKeyPath = CredentialServiceConstants.CLIENT_PUBLIC_KEY_PATH.lookup(getContext());
            String clientSideAlgorithm = CredentialServiceConstants.CLIENT_ENCRYPTION_ALGORITHM.lookup(getContext());

            String errorMessage = "";
            String lineSeparator = System.getProperty("line.separator");

            if (StringUtils.isEmpty(serverPrivateKeyPath)) {
                errorMessage += LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_PROPERTY_IS_EMPTY,
                    CredentialServiceConstants.SERVER_PRIVATE_KEY_PATH.getProperty())
                    + lineSeparator;
            }
            if (StringUtils.isEmpty(serverSideAlgorithm)) {
                errorMessage += LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_PROPERTY_IS_EMPTY,
                    CredentialServiceConstants.SERVER_ENCRYPTION_ALGORITHM.getProperty())
                    + lineSeparator;
            }
            if (StringUtils.isEmpty(clientPublicKeyPath)) {
                errorMessage += LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_PROPERTY_IS_EMPTY,
                    CredentialServiceConstants.CLIENT_PUBLIC_KEY_PATH.getProperty())
                    + lineSeparator;
            }
            if (StringUtils.isEmpty(clientSideAlgorithm)) {
                errorMessage += LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_PROPERTY_IS_EMPTY,
                    CredentialServiceConstants.CLIENT_ENCRYPTION_ALGORITHM.getProperty())
                    + lineSeparator;
            }

            if (!StringUtils.isEmpty(errorMessage)) {
                throw new RuntimeException(errorMessage);
            }

            try {
                setServerPrivateKey(Encryption.getPrivateKeyFromFile(serverPrivateKeyPath, serverSideAlgorithm));
                setServerAlgorithm(serverSideAlgorithm);
            }
            catch (EncryptionException e) {
                throw new RuntimeException(LOGGER.error(
                    MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_LOADING_PRIVATE_KEY, e, serverPrivateKeyPath,
                    serverSideAlgorithm));
            }

            try {
                setClientPublicKey(Encryption.getPublicKeyFromFile(clientPublicKeyPath));
                setClientAlgorithm(clientSideAlgorithm);
            }
            catch (EncryptionException e) {
                throw new RuntimeException(LOGGER.error(
                    MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_LOADING_PUBLIC_KEY, e, clientPublicKeyPath,
                    clientSideAlgorithm));
            }
        }
        else {
            throw new RuntimeException(LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_CONTEXT_NOT_SET));
        }
    }

    public OperationResult addCredential(Credential credential, String type) {
        OperationResult result = new OperationResult();
        result.setState(OperationState.COMPLETE);

        try {
            byte[] hexString = Hex.decodeHex(credential.getPassword().toCharArray());
            byte[] decryptedBytes = Encryption.decrypt(hexString, getServerPrivateKey(), getServerAlgorithm());
            Credential decryptedCredential = new Credential(credential.getKey(), new String(decryptedBytes));

            getCredentialManager().addCredential(decryptedCredential, type);
        }
        catch (CredentialPersistenceException e) {
            LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_ADD_CREDENTIAL, e, credential.getKey(), type,
                e.getMessage());
            errorResult(result, e);
        }
        catch (EncryptionException e) {
            LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_ADD_CREDENTIAL, e, credential.getKey(), type,
                e.getMessage());
            errorResult(result, e);
        }
        catch (DecoderException e) {
            LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_ADD_CREDENTIAL, e, credential.getKey(), type,
                e.getMessage());
            errorResult(result, e);
        }

        return result;
    }

    public Credential getCredential(CredentialKey key, String type) {
        try {
            Credential credential = getCredentialManager().retrieveCredential(key, type);

            byte[] encryptedBytes = Encryption.encrypt(credential.getPassword().getBytes(), getClientPublicKey(),
                getClientAlgorithm());
            return new Credential(credential.getKey(), new String(Hex.encodeHex(encryptedBytes)));
        }
        catch (CredentialPersistenceException e) {
            LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_GET_CREDENTIAL, e, key, type, e.getMessage());
        }
        catch (EncryptionException e) {
            LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_GET_CREDENTIAL, e, key, type, e.getMessage());
        }

        return null;
    }

    public OperationResult removeCredential(CredentialKey key, String type) {
        OperationResult result = new OperationResult();
        result.setState(OperationState.COMPLETE);

        try {
            getCredentialManager().removeCredential(key, type);
        }
        catch (CredentialPersistenceException e) {
            LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_REMOVE_CREDENTIAL, e, key, type, e
                .getMessage());
            errorResult(result, e);
        }

        return result;
    }
    
    /**
     * Currently, listing credentials does not pass the password. If this changes,
     * then the passwords will need to be encrypted exactly the same as getCredential does.
     */
    public List<Credential> listCredentials(String type) {
        try {
            return getCredentialManager().listCredentials(type);
        }
        catch (CredentialPersistenceException e) {
            LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR_LIST_CREDENTIALS, e, type, e.getMessage());
        }
        
        return null;
    }

    public String interfaceVersion() {
        return CredentialServiceConstants.INTERFACE_VERSION.lookup(getContext());
    }

    public String releaseVersion() {
        return CredentialServiceConstants.RELEASE_VERSION.lookup(getContext());
    }

    /**
     * Update the result with the appropriate error message.
     * 
     * @param result
     * @param e
     */
    protected void errorResult(OperationResult result, Exception e) {
        List<LogMessage> messages = new ArrayList<LogMessage>();
        Object[] args = { e.getMessage() };
        messages.add(new LogMessage(LogLevel.ERROR, LocalCredentialService.class.getName(),
            MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_SOAP_ERROR, args));

        if (null != result.getMessages()) {
            result.getMessages().addAll(messages);
        }
        else {
            result.setMessages(messages);
        }

        result.setState(OperationState.FAILED);
    }

    protected ConfigBundle getContext() {
        return context;
    }

    /**
     * This is expected to be set by the caller instantiating this object.
     * 
     * @param config
     */
    public void setContext(ConfigBundle config) {
        this.context = config;
    }

    protected PrivateKey getServerPrivateKey() {
        return serverPrivateKey;
    }

    protected void setServerPrivateKey(PrivateKey serverPrivateKey) {
        this.serverPrivateKey = serverPrivateKey;
    }

    protected CredentialPersistenceManager getCredentialManager() {
        return credentialManager;
    }

    /**
     * This is expected to be set by the caller instantiating this object.
     * 
     * @param credentialManager
     */
    public void setCredentialManager(CredentialPersistenceManager credentialManager) {
        this.credentialManager = credentialManager;
    }

    protected String getServerAlgorithm() {
        return serverAlgorithm;
    }

    protected void setServerAlgorithm(String serverAlgorithm) {
        this.serverAlgorithm = serverAlgorithm;
    }

    protected PublicKey getClientPublicKey() {
        return clientPublicKey;
    }

    protected void setClientPublicKey(PublicKey clientPublicKey) {
        this.clientPublicKey = clientPublicKey;
    }

    protected String getClientAlgorithm() {
        return clientAlgorithm;
    }

    protected void setClientAlgorithm(String clientAlgorithm) {
        this.clientAlgorithm = clientAlgorithm;
    }
}
