/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/EmbeddedDerbyPersistenceManager.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc. 
 * All rights reserved.
 */
package com.netapp.common.flow;

import static com.netapp.common.flow.FlowConstants.PERSISTENCE_EMBEDDED_STORAGE_LOCATION;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import com.netapp.common.MsgKey;
import com.netapp.common.db.derby.DerbyConnectionProperties;
import com.netapp.common.db.derby.EmbeddedDerbyDatabase;
import com.netapp.common.logging.NALogger;

public class EmbeddedDerbyPersistenceManager extends JDBCPersistenceManager {
    private static final NALogger logger = NALogger.getLogger(EmbeddedDerbyPersistenceManager.class);
    private int derbyRetryConnectionCount = 3;
    private EmbeddedDerbyDatabase database;
    private DerbyConnectionProperties connectionProperties;
    
    /**
     * 
     */
    public EmbeddedDerbyPersistenceManager() {
        super(null, Types.BLOB);

        this.database = new EmbeddedDerbyDatabase();
        this.connectionProperties = new DerbyConnectionProperties();
        this.connectionProperties.setCreateDatabase(true);
        this.connectionProperties.setDatabaseName("crashdb");
        
        setDataSource(this.database.getDataSource(this.connectionProperties));
    }

    @Override
    public void init() {
        init(false);
    }

    public void init(boolean forceReload) {
        if (getApplicationContext() != null) {
            this.database.setStorageLocation(PERSISTENCE_EMBEDDED_STORAGE_LOCATION.lookup(
                getApplicationContext()).getAbsolutePath());
        }

        try {
            this.database.startDatabase();
            logger.debug(MsgKey.FLOW_PERSISTENCE_DERBY_STARTED);
        }
        catch (SQLException e) {
            logger.error(MsgKey.FLOW_PERSISTENCE_DERBY_START_FAILED, e, e.getLocalizedMessage());
        }

        if (!validateSchema() || forceReload) {
            reloadSchema();
        }

        super.init();
    }

    @Override
    protected Connection doGetConnection() throws SQLException {
        return this.database.getConnection(this.connectionProperties);
    }

    private boolean validateSchema() {
        try {
            Connection conn = getConnection(this.derbyRetryConnectionCount);
            PreparedStatement ps = null;
            ResultSet rs = null;
            try {
                ps = conn.prepareStatement("SELECT count(*) FROM flow_context");
                rs = ps.executeQuery();
                return true;
            }
            catch (SQLException e) {
                logger.warn(MsgKey.FLOW_PERSISTENCE_DERBY_VALIDATE_FAILED, e.getLocalizedMessage());
            }
            finally {
                closeJdbcResources(rs, ps, conn);
            }
        }
        catch (SQLException e) {
            logger.warn(MsgKey.FLOW_PERSISTENCE_DERBY_VALIDATE_FAILED, e.getLocalizedMessage());
        }

        return false;
    }

    private void reloadSchema() {
        Connection conn = null;
        try {
            conn = getConnection(this.derbyRetryConnectionCount);
            loadFlowContext(conn);
            loadFlowLog(conn);
            loadFlowLock(conn);
        }
        catch (SQLException e) {
            throw new FlowException(logger.warn(MsgKey.FLOW_PERSISTENCE_DERBY_LOAD_FAILED, e, e.getLocalizedMessage()));
        }
        finally {
            closeJdbcResources(null, null, conn);
        }
    }

    private void loadFlowContext(Connection conn) {
        PreparedStatement ps = null;
        try {
            dropTable(conn, "flow_context");

            ps = conn.prepareStatement("CREATE TABLE flow_context (" + "operation_id char(32) primary key,"
                + "client_request_id varchar(255)," + "type varchar(255) not null," + "create_time timestamp not null,"
                + "state varchar(16)," + "hold_until_time timestamp," + "checkpoint_time timestamp,"
                + "checkpoint varchar(255)," + "abort_checkpoint varchar(255)," + "retry_count integer,"
                + "cancel smallint," + "registers blob not null," + "task_exception blob," + "abort_exception blob"
                + ")");
            ps.executeUpdate();
            closeJdbcResources(null, ps, null);

            ps = conn.prepareStatement("CREATE INDEX flow_context_idx1 ON flow_context(client_request_id)");
            ps.executeUpdate();
            closeJdbcResources(null, ps, null);

            ps = conn.prepareStatement("CREATE INDEX flow_context_idx2 ON flow_context(type, hold_until_time)");
            ps.executeUpdate();
            conn.commit();

        }
        catch (SQLException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_PERSISTENCE_DERBY_LOAD_FAILED, e, e.getLocalizedMessage()));
        }
        finally {
            closeJdbcResources(null, ps, null);
        }
    }

    private void dropTable(Connection conn, String tableName) {
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement("DROP TABLE " + tableName);
            ps.executeUpdate();
            conn.commit();
        }
        catch (SQLException e) {
            // ignore
        }
        finally {
            closeJdbcResources(null, ps, null);
        }
    }

    private void loadFlowLog(Connection conn) {
        PreparedStatement ps = null;
        try {
            dropTable(conn, "flow_log");

            ps = conn.prepareStatement("CREATE TABLE flow_log (" + "operation_id char(32) not null,"
                + "log_timestamp timestamp," + "level varchar(16)," + "message varchar(" + getMaxLogMessageLength() + "),"
                + "location varchar(255)," + "key_class varchar(255)," + "key_value varchar(255),"
                + "thread varchar(255)," + "parameters clob" + ")");
            ps.executeUpdate();
            closeJdbcResources(null, ps, null);

            ps = conn.prepareStatement("CREATE INDEX flow_log_idx1 ON flow_log(operation_id, log_timestamp)");
            ps.executeUpdate();
            conn.commit();
        }
        catch (SQLException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_PERSISTENCE_DERBY_LOAD_FAILED, e, e.getLocalizedMessage()));
        }
        finally {
            closeJdbcResources(null, ps, null);
        }
    }

    private void loadFlowLock(Connection conn) {
        PreparedStatement ps = null;
        try {
            dropTable(conn, "flow_lock");

            ps = conn.prepareStatement("CREATE TABLE flow_lock (" + "operation_id char(32) not null,"
                + "lock_key varchar(255) not null," + "primary key(lock_key)" + ")");
            ps.executeUpdate();
            closeJdbcResources(null, ps, null);

            ps = conn.prepareStatement("CREATE INDEX flow_lock_idx1 on flow_lock(operation_id)");
            ps.executeUpdate();
            conn.commit();

        }
        catch (SQLException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_PERSISTENCE_DERBY_LOAD_FAILED, e, e.getLocalizedMessage()));
        }
        finally {
            closeJdbcResources(null, ps, null);
        }
    }

}
