/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/JDBCPersistenceManager.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;

import static com.netapp.common.flow.FlowConstants.CONTEXT_LIFETIME;
import static com.netapp.common.flow.FlowConstants.CONTEXT_REQUEUE_ON_INIT;
import static com.netapp.common.flow.FlowConstants.CONTEXT_SCAVENGE_INCOMPLETE;
import static com.netapp.common.flow.FlowConstants.CONTEXT_SCAVENGE_INTERVAL;
import static com.netapp.common.flow.FlowConstants.PERSISTENCE_CONNECTION_LOG_RETRY_COUNT;
import static com.netapp.common.flow.FlowConstants.PERSISTENCE_CONNECTION_RETRY_COUNT;
import static com.netapp.common.flow.FlowConstants.PERSISTENCE_CONNECTION_RETRY_IDLE;
import static com.netapp.common.flow.FlowConstants.PERSISTENCE_ENCRYPTION_ENABLE;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.sql.DataSource;

import com.netapp.common.MsgKey;
import com.netapp.common.encryption.ByteArrayCipher;
import com.netapp.common.encryption.DES;
import com.netapp.common.flow.api.OperationState;
import com.netapp.common.logging.LogLevel;
import com.netapp.common.logging.LogMessage;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.Cache;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.StringUtils;

/**
 * JDBC Persistence Manager
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class JDBCPersistenceManager implements PersistenceManager {
    private static final NALogger logger = NALogger.getLogger(JDBCPersistenceManager.class);
    private static final String[] STRING_ARRAY_CAST = new String[0];
    private static final Pattern UNPACK_PATTERN = Pattern.compile("(\\\\)?([,\\\\])");
    
    private OperationLogListener logAppender;
    private EngineConfiguration engineConfiguration;
    private DataSource dataSource;
    
    private ConfigBundle applicationContext;
    private ScavengeThread scavengeThread;
    
    private long contextLifetime = 6 * 60;
    private long scavengeInterval = 30;
    private boolean scavengeIncomplete = true;
    private int connectionRetryCount = -1;
    private int connectionRetryIdle = 3000;
    private int connectionLogRetryCount = 5;
    
    private boolean requeueOnInit = true;
    private boolean encryptionEnable;
    private volatile boolean valid;
    
    // Derby uses the type JDBC BLOB for BLOB column and throws an error when using the JDBC type BINARY
    // on a BLOB column.  However, MaxDB needs the type to be BINARY so since Derby is coming in later
    // and is the one not currently working, we will let the Derby Persistence Manager override this value.
    private final int blobJDBCType;
    
    private int maxLogMessageLength = 2048;

    private Cache<String, OperationContext> cache;
    public JDBCPersistenceManager() {
        this(Types.BINARY);
    }

    public JDBCPersistenceManager(int blobJDBCType) {
        this.blobJDBCType = blobJDBCType;
    }

    public JDBCPersistenceManager(DataSource dataSource) {
        this(dataSource, Types.BINARY);
    }
    
    public JDBCPersistenceManager(DataSource dataSource, int blobJDBCType) {
        this.dataSource = dataSource;
        this.blobJDBCType = blobJDBCType;
    }

    @Deprecated
    public void setEngineConfiguration(EngineConfiguration configuration) {
        // This is now handled by start()
    }

    /**
     * Set the datasource pool for this manager.
     * @param dataSource
     */
    public void setDataSource(DataSource dataSource) {
        this.dataSource = dataSource;
    }
    
    protected DataSource getDataSource() {
        return this.dataSource;
    }
    
    public void setApplicationContext(ConfigBundle applicationContext) {
        this.applicationContext = applicationContext;
    }
    
    protected ConfigBundle getApplicationContext() {
        return this.applicationContext;
    }
    
    /** 
     * Init method. To be called after all configuration is complete, generally
     * by Spring or another framework.
     *
     */
    public void init() {
        assert this.dataSource != null;
        if (applicationContext != null) {
            Long value = CONTEXT_LIFETIME.lookup(this.applicationContext);
            this.contextLifetime = (value != null) ? value : this.contextLifetime;
            
            value = CONTEXT_SCAVENGE_INTERVAL.lookup(this.applicationContext);
            this.scavengeInterval = (value != null) ? value : this.scavengeInterval;
            
            Boolean b = CONTEXT_REQUEUE_ON_INIT.lookup(this.applicationContext);
            this.requeueOnInit = (b != null) ? b : this.requeueOnInit;
            
            b = PERSISTENCE_ENCRYPTION_ENABLE.lookup(this.applicationContext);
            this.encryptionEnable = (b != null) ? b : this.encryptionEnable;
            
            b = CONTEXT_SCAVENGE_INCOMPLETE.lookup(this.applicationContext);
            this.scavengeIncomplete = (b != null) ? b : this.scavengeIncomplete;
            
            Integer value2 = PERSISTENCE_CONNECTION_RETRY_COUNT.lookup(this.applicationContext);
            this.connectionRetryCount = (value2 != null) ? value2 : this.connectionRetryCount;
            
            value2 = PERSISTENCE_CONNECTION_RETRY_IDLE.lookup(this.applicationContext);
            this.connectionRetryIdle = (value2 != null) ? value2 : this.connectionRetryIdle;
            
            value2 = PERSISTENCE_CONNECTION_LOG_RETRY_COUNT.lookup(this.applicationContext);
            this.connectionLogRetryCount = (value2 != null) ? value2 : this.connectionLogRetryCount;
        }

        this.logAppender = new OperationLogListener(this, LogLevel.INFO);
        NALogger.addListener(this.logAppender);

        this.cache = new Cache<String, OperationContext>("com.netapp.common.flow:name=JDBCContextCache");
        
        try {
            if (this.requeueOnInit) {
                requeueRunningContexts();
            }
        }
        catch (FlowException e) {
            // We should throw an exception here. However, if we do and we're wiring this together with
            // Spring under a Jetty context (the normal use case), Spring will stop plugging things together
            // but the exception won't bubble up through Jetty and we get a non-working server.
            // See burts 259668, 259670, and 256580.
            // If we blow up on start, we set valid = false and return immediately. The engine will then
            // not start the threadpools to ever make this active.
            // The stack trace will already print on the previous message
            this.valid = false;
            logger.error(MsgKey.FLOW_PERSISTENCE_JDBC_EXCEPTION_DURING_INIT, e.getLocalizedMessage());
            return;
        }
        this.valid = true;
    }
    
    /**
     * To be called by the engine, immediately before it starts handling requests. This gets
     * called after {@link #init()} in the lifecycle.
     * @see com.netapp.common.flow.PersistenceManager#start(Engine)
     */
    public synchronized void start(Engine engine) {
        assert this.engineConfiguration == null;
        this.engineConfiguration = engine.getConfiguration();
        if (this.scavengeThread == null && this.scavengeInterval > 0) {
            this.scavengeThread = new ScavengeThread();
            this.scavengeThread.start();
        }
    }
    
    /**
     * 
     * @see com.netapp.common.flow.PersistenceManager#stop()
     */
    public synchronized void stop() {
        if (this.scavengeThread != null) {
            this.scavengeThread.shutdown();
        }
        this.scavengeThread = null;
        this.engineConfiguration = null;
    }
    
    /**
     * @return SQL Connection
     */
    protected Connection getConnection() throws SQLException {
        return getConnection(connectionRetryCount);
    }
    
    protected Connection getConnection(int retryCount) throws SQLException {
        Connection conn = null;
        int attempt = 0;
        SQLException  sqlEx = null;
        
        while (retryCount == -1 || attempt <= retryCount) {
            try {
                conn = doGetConnection();
                conn.setAutoCommit(false);
                this.valid = true;
                return conn;
            }
            catch (SQLException e) {
                attempt++;
                sqlEx = e;
                this.valid = false;
                logger.error(MsgKey.FLOW_PERSISTENCE_JDBC_UNABLE_TO_GET_CONNECTION, attempt, e.getLocalizedMessage());
                try {
                    Thread.sleep(this.connectionRetryIdle);
                }
                catch (InterruptedException e2) { }
            }
        }
        logger.error(MsgKey.FLOW_PERSISTENCE_JDBC_CONNECTION_RETRY_EXCEEDED, sqlEx, attempt, sqlEx.getLocalizedMessage());
        throw sqlEx;
    }

    protected Connection doGetConnection() throws SQLException {
        Connection conn;
        conn = dataSource.getConnection();
        return conn;
    }
    
    protected final void cacheContext(OperationContext context) {
        this.cache.put(context.getId().getId(), context);
        if (!StringUtils.isEmpty(context.getClientRequestId())) {
            this.cache.put(context.getClientRequestId(), context);
        }
    }
    
    /**
     * 
     * @see com.netapp.common.flow.PersistenceManager#createContext(com.netapp.common.flow.OperationTemplate, java.util.Map)
     */
    public synchronized OperationContext createContext(OperationTemplate template, Map<String, ?> params) {
        OperationContext result = new OperationContext(this, template, params);
        cacheContext(result);
        result.enterContext();
        saveContext(result);
        result.exitContext();
        return result;
    }
    
    /**
     * 
     * @see com.netapp.common.flow.PersistenceManager#getContext(java.lang.String)
     */
    public synchronized OperationContext getContext(String id) {
        Connection conn = null;
        OperationContext result = this.cache.get(id);
        if  (result != null) {
            return result;
        }
        try {
            conn = getConnection();
            result = getContext(conn, id);
            conn.commit();
        }
        catch (SQLException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_PERSISTENCE_JDBC_SQL_EXCEPTION_DURING_LOAD, e), e);
        }
        finally {
            closeJdbcResources(null, null, conn);
        }
        return result;
    }
    
    private OperationContext getContext(Connection conn, String id) throws SQLException {
        PreparedStatement ps = null;
        ResultSet rs = null;
        OperationContext result = null;
        try {
            String whichId = OperationId.isId(id) ? "operation_id" : "client_request_id";
            ps = conn.prepareStatement("SELECT operation_id, client_request_id, type, create_time, state, hold_until_time, "
                + "checkpoint, abort_checkpoint, retry_count, cancel, registers, task_exception, abort_exception"
                + " FROM flow_context WHERE " + whichId + " = ?");
            
            ps.setString(1, id);
            rs = ps.executeQuery();
            if (rs.next()) {
                result = loadOperationContext(rs);
            }
        }
        finally {
            closeJdbcResources(rs, ps, null);
        }
        return result;
    }

    /**
     * Get the next runnable context.
     * The context needs to be set to RUNNING and persisted before returning.
     * @see com.netapp.common.flow.PersistenceManager#getRunnableContext(java.util.Set)
     */
    public synchronized OperationContext getRunnableContext(Set<String> operationTypes) {
        if (operationTypes.isEmpty()) {
            return null;
        }
        
        Connection conn = null;
        PreparedStatement ps = null;
        ResultSet rs = null;
        
        OperationContext result = null;
        
        try {
            conn = getConnection();
            
            // This is somewhat complicated to make sure we don't grab the same thing
            // twice in concurrent attempts without locking anything in the database.
            // We might want to change some of that once we have a target db.
            // Step 1: SELECT for UPDATE a qualifying ID. This will return 0 or 1 ids.
            //         It may return none if two updates tried concurrently against the same row.
            // Step 2: Move the selected row to RUNNING
            // Step 3: Load the selected row
            //
            // UPDATE
            // Getting this to work in a universal way is almost impossible. SQL isn't standard enough
            // to give us a consistient way to lock a row or table. LOCK TABLE syntax and semantics are
            // completely non-standard, MaxDB doesn't support LIMIT X or a suitable analogue combined with
            // ORDER BY.
            // MaxDB requires a column name in conjunction with SELECT ... FOR UPDATE while Postgres
            // takes an optional table name.
            //
            // For this release, we don't attempt database level locking. Each threadpool's scheduler is a
            // single thread, so there should only be one connection modifying that row anyway. To support
            // clustering between multiple workflow servers, we'll require some data model changes and likely
            // DB specific hooks to enforce locking at the DB level.
            StringBuilder idSelect = new StringBuilder();
            Iterator<String> iter = operationTypes.iterator();
            idSelect.append("SELECT operation_id FROM flow_context WHERE type IN ('").append(iter.next()).append("'");
            while (iter.hasNext()) {
                idSelect.append(", '").append(iter.next()).append("'");
            }
            idSelect.append(") AND state = 'PAUSED' AND (hold_until_time is null OR hold_until_time < ?) "
                + " ORDER BY CASE WHEN hold_until_time IS NULL THEN 0 ELSE 1 END, hold_until_time");
            
            String idSelectString = idSelect.toString();
            logger.internal(idSelectString);
            ps = conn.prepareStatement(idSelectString);
            ps.setTimestamp(1, timestamp(new Date()));
            rs = ps.executeQuery();
            
            if (rs.next()) {
                String id = rs.getString("operation_id");
                ps = conn.prepareStatement("UPDATE flow_context SET state = 'RUNNING' WHERE operation_id = ?");
                ps.setString(1, id);
                ps.executeUpdate();
                
                ps = conn.prepareStatement("SELECT operation_id, client_request_id, type, create_time, state, hold_until_time, "
                    + "checkpoint, abort_checkpoint,"
                    + " retry_count, cancel, registers, task_exception, abort_exception"
                    + " FROM flow_context WHERE operation_id = ?");

                ps.setString(1, id);
                rs = ps.executeQuery();
                if (rs.next()) {
                    result = loadOperationContext(rs);
                }
            }
            conn.commit();
        }
        catch (SQLException e) {
            try { 
                conn.rollback(); 
            } 
            catch (SQLException e1) { }
            
            throw new FlowException(logger.error(MsgKey.FLOW_PERSISTENCE_JDBC_SQL_EXCEPTION_DURING_LOAD, e), e);
        }
        finally {
            closeJdbcResources(rs, ps, conn);
        }
        return result;
    }

    /**
     * Move any contexts currently marked as RUNNING into PAUSED
     * and back into the queue to be returned by {@link #getRunnableContext(Set)}.
     * This is executed when the system is brought back up after a crash to recover
     * any contexts in-flight during the crash.
     */
    protected void requeueRunningContexts() {
        logger.debug(MsgKey.FLOW_PERSISTENCE_REQUEUE);
        Connection conn = null;
        Statement stmt = null;
        try {
            conn = getConnection(0);
            stmt = conn.createStatement();
            stmt.executeUpdate("UPDATE flow_context SET state = 'PAUSED', retry_count = retry_count + 1 WHERE state = 'RUNNING'");
            conn.commit();
        }
        catch (SQLException e) {
            if (conn != null) { 
                try { 
                    conn.rollback(); 
                } 
                catch (SQLException e1) { }
                closeJdbcResources(null, null, conn);
            }
            throw new FlowException(
                logger.error(MsgKey.FLOW_PERSISTENCE_JDBC_SQL_EXCEPTION_DURING_REQUEUE, e.getLocalizedMessage()), e);
        }
        finally {
            closeJdbcResources(null, stmt, conn);
        }
    }

    public void saveContext(OperationContext context) {
        logger.debug(MsgKey.FLOW_PERSISTENCE_CHECKPOINT, context.getCheckpoint(), context.getState());
        Connection conn = null;
        PreparedStatement ps = null;
        ResultSet rs = null;
        int attempt = 0;
        // BURT 379489 - This must not fail. Retry forever if it does.
        while (context.isDirty()) {
            try {
                conn = getConnection();
                ps = conn.prepareStatement("SELECT count(*) FROM flow_context WHERE operation_id = ?");
                ps.setString(1, context.getId().getId());
                rs = ps.executeQuery();
                rs.next();

                if (rs.getInt(1) == 0) {
                    insertContext(conn, context);
                }
                else {
                    updateContext(conn, context);
                }
                if (context.isTerminalState(false)) {
                    releaseLocks(conn, context.getId());
                }
                conn.commit();
                context.markClean();
            }
            catch (SQLException e) {
                try {
                    conn.rollback();
                }
                catch (SQLException e2) { }
                attempt += 1;
                logger.error(MsgKey.FLOW_PERSISTENCE_JDBC_SQL_EXCEPTION_DURING_SAVE, e, attempt);
            }
            finally {
                closeJdbcResources(rs, ps, conn);
            }
        }
    }
    
    public synchronized void cancelContext(String id) {
        OperationContext context = this.getContext(id);
        Connection conn = null;
        PreparedStatement ps = null;
        
        if (context == null) {
            return;
        }
        context.enterContext();
        if (!context.isTerminalState()) {
            logger.debug(MsgKey.FLOW_OPERATION_CANCEL_REQUESTED);
            // Set our local copy
            context.cancel();
            // Force it to the DB
            try {
                conn = getConnection();
                ps = conn.prepareStatement("UPDATE flow_context SET cancel = ? WHERE operation_id = ?");
                ps.setBoolean(1, true);
                ps.setString(2, context.getId().getId());
            }
            catch (SQLException e) {
                logger.error(MsgKey.FLOW_PERSISTENCE_JDBC_EXCEPTION_DURING_CANCEL, e, e.getLocalizedMessage());
            }
            finally {
                closeJdbcResources(null, ps, conn);
            }
        }
        else if (context.isAborting()) {
            logger.debug(MsgKey.FLOW_OPERATION_CANNOT_CANCEL_ABORTING_OPERATION);
        }
        else {
            logger.debug(MsgKey.FLOW_OPERATION_CANNOT_CANCEL_FINISHED_OPERATION);
        }
        context.exitContext();
    }
    
    public void releaseContext(String id) {

    }
    
    private void insertContext(Connection conn, OperationContext context) throws SQLException {
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement("INSERT INTO flow_context(operation_id, client_request_id, type, create_time, state,"
                + " hold_until_time, checkpoint_time, checkpoint, abort_checkpoint, retry_count, cancel, registers, task_exception, abort_exception)"
                + " VALUES( ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");

            int c = 1;
            ps.setString(c++, context.getId().getId());
            
            if (!StringUtils.isEmpty(context.getClientRequestId())) {
                ps.setString(c++, context.getClientRequestId());
            }
            else {
                ps.setNull(c++, Types.VARCHAR);
            }
            
            ps.setString(c++, context.getOperationType());
            ps.setTimestamp(c++, new Timestamp(context.getCreateTime().getTime()));
            ps.setString(c++, context.getState().name());

            if (context.getHoldUntilTime() != null) {
                ps.setTimestamp(c++, timestamp(context.getHoldUntilTime()));
            }
            else {
                ps.setNull(c++, Types.TIMESTAMP);
            }

            ps.setTimestamp(c++, timestamp(context.getCheckpointTime()));

            if (context.getCheckpoint() != null) {
                ps.setString(c++, context.getCheckpoint());
            }
            else {
                ps.setNull(c++, Types.VARCHAR);
            }

            if (context.getAbortCheckpoint() != null) {
                ps.setString(c++, context.getAbortCheckpoint());
            }
            else {
                ps.setNull(c++, Types.VARCHAR);
            }

            ps.setInt(c++, context.getRetryCount());

            ps.setBoolean(c++, context.isCancelled());

            byte[] registers = serializeObject(context.getPersistentRegisters());
            
            if (this.encryptionEnable) {
                ByteArrayCipher des = new DES(context.getId().getId());
                registers = des.encrypt(registers);
            }
            ps.setBytes(c++, registers);

            if (context.getTaskException() != null) {
                ps.setBytes(c++, serializeObject(context.getTaskException()));
            }
            else {
                ps.setNull(c++, this.blobJDBCType);
            }

            if (context.getAbortException() != null) {
                ps.setBytes(c++, serializeObject(context.getAbortException()));
            }
            else {
                ps.setNull(c++, this.blobJDBCType);
            }

            ps.executeUpdate();
        }
        finally {
            closeJdbcResources(null, ps, null);
        }
    }
    
    private void updateContext(Connection conn, OperationContext context) throws SQLException {
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement("UPDATE flow_context SET"
                + " state = ?, hold_until_time = ?, checkpoint_time = ?, checkpoint = ?, abort_checkpoint = ?,"
                + " retry_count = ?, cancel = ?, registers = ?, task_exception = ?, abort_exception = ?"
                + " WHERE operation_id = ?");

            int c = 1;

            ps.setString(c++, context.getState().toString());

            if (context.getHoldUntilTime() != null) {
                ps.setTimestamp(c++, timestamp(context.getHoldUntilTime()));
            }
            else {
                ps.setNull(c++, Types.TIMESTAMP);
            }

            ps.setTimestamp(c++, timestamp(context.getCheckpointTime()));

            if (context.getCheckpoint() != null) {
                ps.setString(c++, context.getCheckpoint());
            }
            else {
                ps.setNull(c++, Types.VARCHAR);
            }

            if (context.getAbortCheckpoint() != null) {
                ps.setString(c++, context.getAbortCheckpoint());
            }
            else {
                ps.setNull(c++, Types.VARCHAR);
            }

            ps.setInt(c++, context.getRetryCount());
            ps.setBoolean(c++, context.isCancelled());

            byte[] registers = serializeObject(context.getPersistentRegisters());
            if (this.encryptionEnable) {
                ByteArrayCipher des = new DES(context.getId().getId());
                registers = des.encrypt(registers);
            }
            ps.setBytes(c++, registers);

            if (context.getTaskException() != null) {
                ps.setBytes(c++, serializeObject(context.getTaskException()));
            }
            else {
                ps.setNull(c++, this.blobJDBCType);
            }

            if (context.getAbortException() != null) {
                ps.setBytes(c++, serializeObject(context.getAbortException()));
            }
            else {
                ps.setNull(c++, this.blobJDBCType);
            }

            ps.setString(c++, context.getId().getId());
            ps.executeUpdate();
        }
        finally {
            closeJdbcResources(null, ps, null);
        }
    }

    @SuppressWarnings("unchecked") // Cast on deserialize of Registers
    private OperationContext loadOperationContext(ResultSet rs) throws SQLException {
        OperationId id = new OperationId(rs.getString("operation_id"));
        String operationType = rs.getString("type");
        
        OperationTemplate template = engineConfiguration.getOperationTemplate(operationType);
        if (template == null) {
            logger.error(MsgKey.FLOW_PERSISTENCE_OPERATION_NOT_FOUND, operationType, id);
            return null;
        }
        
        OperationContext result = this.cache.get(id.getId());
        if (result == null) {
            result = new OperationContext(this, id, template);
            result.setClientRequestId(rs.getString("client_request_id"));
            result.setCreateTime(date(rs.getTimestamp("create_time")));
        }
        
        OperationState state = OperationState.valueOf(rs.getString("state"));
        result.setState(state);
        
        result.setHoldUntilTime(date(rs.getTimestamp("hold_until_time")));
        result.setCheckpoint(rs.getString("checkpoint"));
        result.setAbortCheckpoint(rs.getString("abort_checkpoint"));
        result.setRetryCount(rs.getInt("retry_count"));
        // We never want to move a canceled context back to not-canceled
        // In the impossible event that the in-memory context is canceled but
        // the database isn't, the memory copy wins.
        result.setCancel(result.isCancelled() | rs.getBoolean("cancel"));
        byte[] bytes = rs.getBytes("registers");
        // TODO: This should probably be stored in the row instead of depending
        // on the current state.
        if (this.encryptionEnable) {
            ByteArrayCipher des = new DES(id.getId());
            bytes = des.decrypt(bytes);
        }
        result.setRegisters((Map<String, Object>)deserializeObject(bytes));
        bytes = rs.getBytes("task_exception");
        if (bytes != null) {
            result.setTaskException((Throwable)deserializeObject(bytes));
        }
        bytes = rs.getBytes("abort_exception");
        if (bytes != null) {
            result.setAbortException((Throwable)deserializeObject(bytes));
        }
        result.markClean();
        cacheContext(result);
        return result;
    }
    
    /**
     * Try to serialize an exception. If that fails, serialize a base exception with the same message.
     * This is to patch around people who make non-serializable exceptions.
     * @param t
     * @return Serialized t
     */
    protected byte[] serializeObject(Throwable t) {
        try {
            return serializeObject((Object)t);
        }
        catch (FlowException e) {
            // Create a new exception with the message and stack trace of the old one
            // Any other details attached are lost.
            Exception e2 = new Exception(t.getClass().getName() + ": " + t.getMessage());
            e2.setStackTrace(t.getStackTrace());
            return serializeObject(e2);
        }
    }
    
    protected byte[] serializeObject(Object obj) {
        ByteArrayOutputStream stream = new ByteArrayOutputStream();
        ObjectOutputStream objStream;
        
        try {
            objStream = new ObjectOutputStream(stream);
            objStream.writeObject(obj);
        }
        catch (IOException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_PERSISTENCE_SERIALIZATION_EXCEPTION, e), e);
        }  
        return stream.toByteArray();
    }
    
    protected Object deserializeObject(byte[] bytes) {
        Object result;
        ByteArrayInputStream stream = new ByteArrayInputStream(bytes);
        try {
            ObjectInputStream objStream = new ObjectInputStream(stream);
            result = objStream.readObject();
        }
        catch (IOException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_PERSISTENCE_SERIALIZATION_EXCEPTION, e), e);
        }
        catch (ClassNotFoundException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_PERSISTENCE_SERIALIZATION_EXCEPTION, e), e);
        } 
        return result;
    }
    
    /**
     * Pack an array of strings into a single string, escaping as needed to split
     * apart later.
     * @param in 
     * @return String representation of in
     */
    protected String pack(List<String> in) {
        if (in == null) {
            return null;
        }
        StringBuilder out = new StringBuilder();
        boolean first = true;
        for (String entry : in) {
            if (!first) {
                out.append(",");
            }
            out.append(entry.replaceAll("([\\\\,])", "\\\\$1"));
            first = false;
        }
        return out.toString();
    }

    /**
     * Break a string packed by pack back into an array.
     * @param in
     * @return List of strings from in
     */
    protected List<String> unpack(String in) {
        if (in == null) {
            return null;
        }
        List<String> out = new ArrayList<String>();
        Matcher m = UNPACK_PATTERN.matcher(in);
        StringBuffer sb = new StringBuffer();
        while (m.find()) {
            if (m.group(2).equals(",") && m.group(1) == null) {
                m.appendReplacement(sb, "");
                out.add(sb.toString());
                sb.setLength(0);
            }
            else {
                m.appendReplacement(sb, "$2");
            }
        }
        m.appendTail(sb);
        out.add(sb.toString());
        return out;
    }
    
    private static Timestamp timestamp(Date date) {
        if (date != null) {
            return new Timestamp(date.getTime());
        }
        return null;
    }
    
    private static Date date(Timestamp timestamp) {
        if (timestamp != null) {
            return new Date(timestamp.getTime());
        }
        return null;
    }
    
    protected static void closeJdbcResources(ResultSet rs, Statement stmt, Connection conn) {
        if (rs != null) {
            try { 
                rs.close();
            } 
            catch (SQLException e2) { }
        }
        
        if (stmt != null) { 
            try { 
                stmt.close(); 
            } 
            catch (SQLException e2) { } 
        }
        
        if (conn != null) {
            try { 
                conn.close(); 
            } 
            catch (SQLException e2) { }
        }
    }

    public void addMessage(OperationId id, LogMessage message) {
        if (message.getLocation() == JDBCPersistenceManager.class.getName()) {
            // We don't log our own messages.
            // If we did, we'd get into a vicious cycle when things went wrong
            // This should be adjusted if we ever descend from this class.
            // Ideally, if we remove the OperationContext (burt 258039) we would
            // only log to the OperationContext structure in memory in the working thread
            // and log to the database in an asynchronous worker thread that didn't have
            // the Operation MDC setup. That way, any errors in anything we called from
            // here wouldn't get logged.
            return;
        }
        
        Connection conn = null;
        PreparedStatement ps = null;
        
        try {
            conn = this.getConnection(connectionLogRetryCount);
            ps = conn.prepareStatement("INSERT INTO flow_log (operation_id, log_timestamp, level, message, "
                + "location, key_class, key_value, thread, parameters) " 
                + "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?)");
            int c = 1;
            ps.setString(c++, id.toString());
            ps.setTimestamp(c++, timestamp(message.getTimestamp()));
            ps.setString(c++, message.getLevel().name());
            ps.setString(c++, StringUtils.truncate(message.getMessage(), getMaxLogMessageLength()));
            ps.setString(c++, message.getLocation());
            ps.setString(c++, message.getMsgKeyClass());
            ps.setString(c++, message.getMsgKeyValue());
            ps.setString(c++, message.getThread());
            String packedParameters = pack(message.getParameters());
            if (packedParameters != null) {
                ps.setString(c++, pack(message.getParameters()));
            }
            else {
                ps.setNull(c++, Types.VARCHAR);
            }
            ps.executeUpdate();
            conn.commit();
        }
        catch (SQLException e) {
            // This is only allowed because we will not log it 
            logger.error(MsgKey.FLOW_PERSISTENCE_JDBC_UNABLE_TO_PERSIST_LOG, e, e.getLocalizedMessage());
        }
        finally {
            closeJdbcResources(null, ps, conn);
        }
        
    }
    

    public List<LogMessage> getMessages(OperationId id) {
        Connection conn = null;
        PreparedStatement ps = null;
        ResultSet rs = null;
        List<LogMessage> results = new ArrayList<LogMessage>();
        
        try {
            conn = this.getConnection();
            ps = conn.prepareStatement("SELECT level, log_timestamp, message, location, key_class, key_value, thread, parameters"
                + " FROM flow_log WHERE operation_id = ? ORDER BY log_timestamp");
            ps.setString(1, id.toString());
            rs = ps.executeQuery();
            while (rs.next()) {
                LogMessage message = new LogMessage();
                message.setLevel(LogLevel.parseLevel(rs.getString("level")));
                message.setTimestamp(date(rs.getTimestamp("log_timestamp")));
                message.setMessage(rs.getString("message"));
                message.setLocation(rs.getString("location"));
                message.setMsgKeyClass(rs.getString("key_class"));
                message.setMsgKeyValue(rs.getString("key_value"));
                message.setThread(rs.getString("thread"));
                message.setParameters(unpack(rs.getString("parameters")));
                results.add(message);
            }
        }
        catch (SQLException e) {
            e.printStackTrace();
        }
        finally {
            closeJdbcResources(rs, ps, conn);
        }
        // TODO Auto-generated method stub
        return results;
    }

    public synchronized void acquireLocks(OperationId id, Set<String> keys) throws UnableToAcquireLockException {
        Connection conn = null;
        PreparedStatement ps = null;
        ResultSet rs = null;
        
        // Keep the set of keys we don't already own
        Set<String> newKeys = new HashSet<String>();
        
        try {
            conn = this.getConnection();
            
            ps = conn.prepareStatement("SELECT fl.operation_id, fc.client_request_id, fc.type FROM "
                + "flow_lock fl left outer join flow_context fc on (fl.operation_id = fc.operation_id) "
                + "where fl.lock_key = ?");
            for (String key : keys) {
                /* 
                 * Just do a join to get the data we need for the error message
                 * We don't need the overhead of creating the full context, and if we try, we hit
                 * a race condition that can stomp on the current cache. See burt 285072
                 * This join syntax works on Derby, MaxDB and PostgreSQL. Oracle may not work.
                 * We do an outer join here to catch the pathological case where the lock exists but the
                 * context doesn't. (Why isn't that a FK constraint???) -jrs
                 */

                ps.setString(1, key);
                rs = ps.executeQuery();
                if (rs.next()) {
                    String holderId = rs.getString(1);
                    String holderClientId = rs.getString(2);
                    String holderType = rs.getString(3);
                    if (!id.getId().equals(holderId)) {
                        if (holderType != null) {
                            throw new UnableToAcquireLockException(new OperationId(holderId), keys,
                                logger.error(MsgKey.FLOW_PERSISTENCE_LOCK_IN_USE, key, 
                                    holderType, 
                                    holderId, 
                                    StringUtils.isEmpty(holderClientId) ? "" : holderClientId));
                        }
                        else {
                            // We have a lock, but the operation that held it seems to be gone.
                            // This really shouldn't happen unless someone touched the database underneath us.
                            // Clear out the locks for that operation and log a warning.
                            logger.warn(MsgKey.FLOW_PERSISTENCE_LOCK_ABANDONED, key, holderId);
                            releaseLocks(conn, new OperationId(holderId));
                            newKeys.add(key);
                        }
                    }
                }
                else {
                    newKeys.add(key);
                }
            }
            closeJdbcResources(rs, ps, null);
            rs = null;
            ps = null;
            
            ps = conn.prepareStatement("INSERT INTO flow_lock(operation_id, lock_key) VALUES(?, ?)");
            for (String key : newKeys) {
                logger.internal("Accquiring lock %s", key);
                ps.setString(1, id.getId());
                ps.setString(2, key);
                ps.executeUpdate();
            }
            closeJdbcResources(null, ps, null);
            ps = null;
            
            conn.commit();
        }
        catch (SQLException ex) {
            if (conn != null) {
                try {
                    conn.rollback();
                }
                catch (SQLException e) {
                    logger.error(MsgKey.FLOW_PERSISTENCE_JDBC_EXCEPTION_DURING_ROLLBACK, e, e.getLocalizedMessage());
                }
            }
            throw new UnableToAcquireLockException(null, keys,
                logger.error(MsgKey.FLOW_PERSISTENCE_SQL_EXCEPTION_ACQUIRING_LOCK, ex, keys, ex.getLocalizedMessage()), ex);
        }
        finally {
            closeJdbcResources(rs, ps, conn);
        }
    }

    protected void releaseLocks(Connection conn, OperationId id) throws SQLException {
        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement("DELETE FROM flow_lock WHERE operation_id = ?");
            ps.setString(1, id.getId());
            ps.executeUpdate();
        }
        finally {
            closeJdbcResources(null, ps, null);
        }
    }
    
    public boolean isValid() {
        return this.valid;
    }
    
    public int getMaxLogMessageLength() {
        return maxLogMessageLength;
    }

    public void setMaxLogMessageLength(int maxLogMessageLength) {
        this.maxLogMessageLength = maxLogMessageLength;
    }
    
    private class ScavengeThread extends Thread {
        private final String[] scrubTables = { "flow_lock", "flow_log", "flow_context" };
        private boolean shutdown;
        
        private ScavengeThread() {
            this.setName("JDBC Context Scavenge Thread");
            this.setDaemon(true);
        }
        
        private synchronized void shutdown() {
            this.shutdown = true;
            this.notify();
        }
        
        public void run() {
            while (!shutdown) {
                Connection conn = null;
                PreparedStatement ps = null;
                ResultSet rs = null;
                if (JDBCPersistenceManager.this.dataSource != null) {
                    try {
                        // MaxDB doesn't appear to have a sane SELECT INTO TEMP TABLE statement
                        Timestamp date = new Timestamp(System.currentTimeMillis() - (contextLifetime * 60 * 1000));
                        logger.debug(MsgKey.FLOW_PERSISTENCE_JDBC_SCAVENGE, date);
                        conn = getConnection(0);
                        ps = conn.prepareStatement("SELECT operation_id FROM flow_context WHERE state IN ('COMPLETE', 'FAILED'"
                            + (scavengeIncomplete ? ", 'INCOMPLETE'" : "")
                            + ") AND checkpoint_time < ?");
                        ps.setTimestamp(1, date);
                        rs = ps.executeQuery();
                        
                        List<String> ids = new ArrayList<String>();
                        while (rs.next()) {
                            ids.add(rs.getString(1));
                        }
                        closeJdbcResources(rs, ps, null);
                        rs = null;
                        
                        for (String table : scrubTables) {
                            ps = conn.prepareStatement("DELETE FROM " + table + " WHERE operation_id = ?");
                            for (String s : ids) {
                                ps.setString(1, s);
                                ps.executeUpdate();
                            }
                            closeJdbcResources(null, ps, null);
                        }
                        conn.commit();
                        synchronized (JDBCPersistenceManager.this) {
                            for (String id : ids) {
                                OperationContext ctx = JDBCPersistenceManager.this.cache.get(id);
                                if (ctx != null) {
                                    JDBCPersistenceManager.this.cache.remove(id);
                                    if (!StringUtils.isEmpty(ctx.getClientRequestId())) {
                                        JDBCPersistenceManager.this.cache.remove(ctx.getClientRequestId());
                                    }
                                }
                            }
                        }
                    }
                    catch (SQLException e) {
                        logger.warn(MsgKey.FLOW_PERSISTENCE_JDBC_SQL_EXCEPTION_DURING_SCAVENGE, e, e.getLocalizedMessage());
                    }
                    finally {
                        closeJdbcResources(rs, ps, conn);
                    }
                }
                try {
                    Thread.sleep(scavengeInterval * 60 * 1000);
                }
                catch (InterruptedException e) { }
            }
        }
    }
}
