/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/OperationTemplate.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;


import static com.netapp.common.flow.FlowConstants.ABORT_POINT_ATTRIBUTE;
import static com.netapp.common.flow.FlowConstants.ABORT_TASKS_ELEMENT;
import static com.netapp.common.flow.FlowConstants.CHECKPOINT_ELEMENT;
import static com.netapp.common.flow.FlowConstants.FOR_ELEMENT;
import static com.netapp.common.flow.FlowConstants.NAME_ATTRIBUTE;
import static com.netapp.common.flow.FlowConstants.OPERATION_ELEMENT;
import static com.netapp.common.flow.FlowConstants.WORKFLOW_SCHEMA_NAMESPACE;
import static com.netapp.common.flow.FlowConstants.REGISTER_ELEMENT;
import static com.netapp.common.flow.FlowConstants.TASKS_ELEMENT;
import static com.netapp.common.flow.FlowConstants.THREADPOOL_ATTRIBUTE;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Node;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.xml.DOMUtils;
/**
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
class OperationTemplate implements StackFrameTemplate {
    private static final NALogger logger = NALogger.getLogger(OperationTemplate.class);
    private String threadpool;
    private String name;
    
    protected Map<String, RegisterTemplate> registerVars = new HashMap<String, RegisterTemplate>();;
    
    protected List<ExecutionStep> taskSteps;
    protected List<ExecutionStep> abortSteps;
    private TaskInstanceFactory taskInstanceFactory;
    
    OperationTemplate(String name, Node node, TaskInstanceFactory taskInstanceFactory) {
        this.name = name;
        this.taskInstanceFactory = taskInstanceFactory;
        
        if (!DOMUtils.isElement(node, WORKFLOW_SCHEMA_NAMESPACE, OPERATION_ELEMENT)) {
            throw new FlowException(logger.error(MsgKey.FLOW_CONFIGURATION_ROOT_ELEMENT_NOT_OPERATION, node.getLocalName()));
        }

        this.threadpool = DOMUtils.getAttribute(node, THREADPOOL_ATTRIBUTE);
        
        Node taskStepsNode = null;
        Node abortStepsNode = null;
        for (Node child = node.getFirstChild(); child != null; child = child.getNextSibling()) {
            if (Node.ELEMENT_NODE == child.getNodeType() && WORKFLOW_SCHEMA_NAMESPACE.equals(child.getNamespaceURI())) {
                if (REGISTER_ELEMENT.equals(child.getLocalName())) {
                    loadRegister(child);
                }
                else if (TASKS_ELEMENT.equals(child.getLocalName())) {
                    assert taskStepsNode == null; // XML schema should validate this
                    taskStepsNode = child;
                }
                else if (ABORT_TASKS_ELEMENT.equals(child.getLocalName())) {
                    assert abortStepsNode == null;
                    abortStepsNode = child;
                }
            }
        }
        this.taskSteps = loadSteps(taskStepsNode);
        this.abortSteps = loadSteps(abortStepsNode);
        validate();
    }
    
    OperationTemplate(String name, String threadpool, TaskInstanceFactory taskInstanceFactory) {
        this.name = name;
        this.threadpool = threadpool;
        this.taskInstanceFactory = taskInstanceFactory;
    }
    
    public String getThreadpool() {
        return this.threadpool;
    }
    
    public List<ExecutionStep> getTaskSteps() {
        return this.taskSteps;
    }
    
    public List<ExecutionStep> getAbortSteps() {
        return this.abortSteps;
    }
    
    /**
     * @see com.netapp.common.flow.StackFrameTemplate#getRegisterInstanceTemplate(java.lang.String)
     */
    public RegisterTemplate getRegisterInstanceTemplate(String name) {
        return registerVars.get(name);
    }
    
    /**
     * Return a public information structure about this operation.
     * @return Information structure describing the operation
     */
    public OperationInformation getInformation() {
        OperationInformation info = new OperationInformation();
        for (RegisterTemplate register : this.registerVars.values()) {
            if (register.isInput()) {
                info.getInputs().put(register.getInputName(), register.getType());
            }
            
            if (register.isOutput()) {
                info.getOutputs().put(register.getOutputName(), register.getType());
            }
        }
        return info;
    }
    
    /**
     * Return the set of input variables for this operation.
     * This exposes the implementation that callers outside this
     * package shouldn't see.
     * @return Set of input register templates
     */
    protected Set<RegisterTemplate> getInputs() {
        Set<RegisterTemplate> inputs = new HashSet<RegisterTemplate>();
        for (RegisterTemplate register : this.registerVars.values()) {
            if (register.isInput()) {
                inputs.add(register);
            }
        }
        return inputs;
    }
    
    protected Set<RegisterTemplate> getOutputs() {
        Set<RegisterTemplate> outputs = new HashSet<RegisterTemplate>();
        for (RegisterTemplate register : this.registerVars.values()) {
            if (register.isOutput()) {
                outputs.add(register);
            }
        }
        return outputs;
    }
    
    protected void loadRegister(Node node) {
        RegisterTemplate template = new RegisterTemplate(node);
        if (this.registerVars.containsKey(template.getName())) {
            throw new FlowException(logger.error(MsgKey.FLOW_REGISTER_REDEFINED, template.getName()));
        }
        this.registerVars.put(template.getName(), template);
        logger.internal("Defining register %s", template.toString());
    }
    
    protected List<ExecutionStep> loadSteps(Node taskList) {
        List<ExecutionStep> stepList = new ArrayList<ExecutionStep>();
        
        if (taskList == null) {
            return stepList;
        }
        
        for (Node child = taskList.getFirstChild(); child != null; child = child.getNextSibling()) {
            if (Node.ELEMENT_NODE != child.getNodeType()) {
                continue;
            }
            ExecutionStep step = loadExecutionStep(child);
            if (step != null) {
                stepList.add(step);
            }
        }
        return stepList;
    }
    
    protected ExecutionStep loadExecutionStep(Node node) {
        ExecutionStep step = null;
        if (DOMUtils.isElement(node, WORKFLOW_SCHEMA_NAMESPACE, CHECKPOINT_ELEMENT)) {
            String stepName = DOMUtils.getAttribute(node, NAME_ATTRIBUTE);
            String abortPoint = DOMUtils.getAttribute(node, ABORT_POINT_ATTRIBUTE);
            step = new CheckpointTemplate(stepName, abortPoint);
        }
        else if (DOMUtils.isElement(node, WORKFLOW_SCHEMA_NAMESPACE, FOR_ELEMENT)) {
            step = new ForLoopTemplate(this, node, this.taskInstanceFactory);
        }
        else if (Node.ELEMENT_NODE == node.getNodeType()) {
            step = new TaskInstanceTemplate(this, node, this.taskInstanceFactory);
        }
        return step;
    }
    
    protected void validate() {
        Set<String> checkpoints = new HashSet<String>();
        Set<String> abortPoints = new HashSet<String>();
        
        for (ExecutionStep step : this.abortSteps) {
            if (step instanceof CheckpointTemplate) {
                CheckpointTemplate checkpoint = (CheckpointTemplate)step;
                if (abortPoints.contains(checkpoint.getName())) {
                    throw new FlowException(logger.error(MsgKey.FLOW_ABORT_CHECKPOINT_REDEFINED, checkpoint.getName(), this.name));
                }
                abortPoints.add(checkpoint.getName());
            }
        }
        
        for (ExecutionStep step : this.taskSteps) {
            if (step instanceof CheckpointTemplate) {
                CheckpointTemplate checkpoint = (CheckpointTemplate)step;
                if (checkpoint.getName() != null && checkpoints.contains(checkpoint.getName())) {
                    throw new FlowException(logger.error(MsgKey.FLOW_CHECKPOINT_REDEFINED, checkpoint.getName(), this.name));   
                }
                if (checkpoint.hasAbortPoint() && !(abortPoints.contains(checkpoint.getAbortPoint()))) {
                    throw new FlowException(logger.error(MsgKey.FLOW_ABORT_CHECKPOINT_NOT_FOUND, checkpoint.getAbortPoint(), this.name));
                }
                checkpoints.add(checkpoint.getName());
            }
        }
    }
    
    public String getName() {
        return this.name;
    }
    
    /**
     * Convinience method for validation
     */
    protected boolean checkpointExists(String name) {
        for (ExecutionStep step : this.taskSteps) {
            if ((step instanceof CheckpointTemplate) 
                    && ((CheckpointTemplate)step).getName().equals(name)) {
                return true;
            }
        }
        return false;
    }
    
    protected boolean abortCheckpointExists(String name) {
        for (ExecutionStep step : this.abortSteps) {
            if ((step instanceof CheckpointTemplate)
                    && ((CheckpointTemplate)step).getName().equals(name)) {
                return true;
            }
        }
        return false;
    }
}
