/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/TaskUtils.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;

import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.CollectionUtils;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.ConfigProperty;

/**
 * Static methods for assisting Tasks and accessing information in the thread context (which is
 * not normally exposed to a running Task).
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class TaskUtils {
    private static final NALogger logger = NALogger.getLogger(TaskUtils.class);

    private static final ThreadLocal<Engine> THREAD_ENGINE = new ThreadLocal<Engine>();
    
    private TaskUtils() {
        assert false;
    }
    /**
     * Retrieve the operation id of the operation running in the current thread.
     * @return OperationId for the running thread
     * @throws IllegalThreadStateException The current running thread does not have a operation context set
     */
    public static OperationId currentOperationId() {
        assertThreadState();
        return OperationContext.getThreadContext().getId();
    }
    
    /**
     * Retrieve the client request id of the operation running in the current thread
     * @return client request id for the running thread
     * @throws IllegalThreadStateException The current running thread does not have a operation context set
     */
    public static String currentClientRequestId() {        
        assertThreadState();
        return OperationContext.getThreadContext().getClientRequestId();
    }

    public static Date currentStartTime() {
        assertThreadState();
        return OperationContext.getThreadContext().getCreateTime();
    }
    
    /**
     * Locks are managed by the {@link PersistenceManager} to prevent concurrent operations
     * from working on the same external objects. Locks are defined by arbitrary string keys
     * describing the external object. For example, all operations acting on a filer volume may
     * aggree to lock keys such as "Volume:filer:/vol/volumeName".
     * Once an operation {@link #acquireLock(Object) acquires a lock}, it is held until
     * the end of the operation. Any subsequent operations attempting to acquire a lock on the
     * same key will result in a {@link UnableToAcquireLockException} exception being thrown.
     * <p>
     * Acquire a lock for the operation running in the current thread on the supplied key.
     * If key is a String, the string itself will be locked {@link #acquireLock(String)},
     * otherwise a unique lock name will be generated based on the key class and toString value.
     * 
     * @param key Key value for the requested lock
     * @throws UnableToAcquireLockException 
     * @throws IllegalThreadStateException The current running thread does not have a operation context set
     *
     */
    public static void acquireLock(Object key) throws UnableToAcquireLockException {
        if (key instanceof String) {
            acquireLock((String)key);
        }
        else {
            acquireLock((String)(key.getClass().getName() + ":" + key.toString()));
        }
    }
    
    /**
     * @param key Key value for the requested lock
     * @throws UnableToAcquireLockException 
     * @see #acquireLock(Object)
     * @throws IllegalThreadStateException The current running thread does not have a operation context set
     */
    public static void acquireLock(String key) throws UnableToAcquireLockException {
        assertThreadState();
        OperationContext.getThreadContext().acquireLocks(CollectionUtils.addAll(new HashSet<String>(), key));
    }

    /**
     * Locks are managed by the {@link PersistenceManager} to prevent concurrent operations
     * from working on the same external objects. Locks are defined by arbitrary string keys
     * describing the external object. For example, all operations acting on a filer volume may
     * aggree to lock keys such as "Volume:filer:/vol/volumeName".
     * Once an operation {@link #acquireLock(Object) acquires a lock}, it is held until
     * the end of the operation. Any subsequent operations attempting to acquire a lock on the
     * same key will result in a {@link UnableToAcquireLockException} exception being thrown.
     * <p>
     * Acquire locks for the operation running in the current thread on the supplied key.
     * If a key is a String, the string itself will be locked {@link #acquireLock(String)},
     * otherwise a unique lock name will be generated based on the key class and toString value.
     * 
     * @param keys Key values for the requested locks
     * @throws UnableToAcquireLockException 
     * @throws IllegalThreadStateException The current running thread does not have a operation context set
     *
     */
    public static void acquireLocks(Object...keys) throws UnableToAcquireLockException {
        Set<String> lockKeys = new HashSet<String>();
        
        for (Object key : keys) {
            if (key instanceof String) {
                lockKeys.add((String)key);
            }
            else {
                lockKeys.add((String)(key.getClass().getName() + ":" + key.toString()));
            }
        }
        
        acquireLocksInternal(lockKeys);
    }
    
    /**
     * Locks are managed by the {@link PersistenceManager} to prevent concurrent operations
     * from working on the same external objects. Locks are defined by arbitrary string keys
     * describing the external object. For example, all operations acting on a filer volume may
     * aggree to lock keys such as "Volume:filer:/vol/volumeName".
     * Once an operation {@link #acquireLock(Object) acquires a lock}, it is held until
     * the end of the operation. Any subsequent operations attempting to acquire a lock on the
     * same key will result in a {@link UnableToAcquireLockException} exception being thrown.
     * <p>
     * Acquire locks for the operation running in the current thread on the supplied key.
     * If a key is a String, the string itself will be locked {@link #acquireLock(String)},
     * otherwise a unique lock name will be generated based on the key class and toString value.
     * 
     * @param keys Key values for the requested locks
     * @throws UnableToAcquireLockException 
     * @throws IllegalThreadStateException The current running thread does not have a operation context set
     *
     */
    public static void acquireLocks(Collection<?> keys) throws UnableToAcquireLockException {
        Set<String> lockKeys = new HashSet<String>();
        
        for (Object key : keys) {
            if (key instanceof String) {
                lockKeys.add((String)key);
            }
            else {
                lockKeys.add((String)(key.getClass().getName() + ":" + key.toString()));
            }
        }
        
        acquireLocksInternal(lockKeys);
    }
    

    /**
     * @param keys Key values for the requested locks
     * @throws UnableToAcquireLockException 
     * @see #acquireLocks(Collection)
     * @throws IllegalThreadStateException The current running thread does not have a operation context set
     */
    public static void acquireLocks(String...keys) throws UnableToAcquireLockException {
        assertThreadState();
        OperationContext.getThreadContext().acquireLocks(CollectionUtils.addAll(new HashSet<String>(), keys));
    }

    /**
     * @param keys Key values for the requested locks
     * @throws UnableToAcquireLockException 
     * @see #acquireLocks(Collection)
     * @throws IllegalThreadStateException The current running thread does not have a operation context set
     */
    private static void acquireLocksInternal(Set<String> keys) throws UnableToAcquireLockException {
        assertThreadState();
        OperationContext.getThreadContext().acquireLocks(keys);
    }

    static void setThreadEngine(Engine threadEngine) {
        TaskUtils.THREAD_ENGINE.set(threadEngine);
    }
    
    public static <T> T configValue(ConfigProperty<T> config) {
        return config.lookup(getApplicationContext());
    }
    
    public static <T> T configValue(ConfigProperty<T> config, T defaultValue) {
        T value = configValue(config);
        return value != null ? value : defaultValue;
    }
    
    public static ConfigBundle getApplicationContext() {
        assertThreadState();
        if (TaskUtils.THREAD_ENGINE.get().getApplicationContext() == null) {
            throw new IllegalThreadStateException(logger.error(MsgKey.FLOW_OPERATION_UTIL_APPLICATION_CONTEXT_NOT_DEFINED));
        }
        return TaskUtils.THREAD_ENGINE.get().getApplicationContext();
    }
    
    private static void assertThreadState() {
        if (TaskUtils.THREAD_ENGINE.get() == null) {
            throw new IllegalThreadStateException(logger.error(MsgKey.FLOW_OPERATION_UTIL_INVALID_THREAD));
        }
    }
}
