/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/soap/WorkflowBindingFactory.java#1 $
 * 
 * Copyright (c) 2008 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow.soap;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.MethodInterceptor;
import net.sf.cglib.proxy.MethodProxy;

import com.netapp.common.MsgKey;
import com.netapp.common.flow.Engine;
import com.netapp.common.flow.EngineConfiguration;
import com.netapp.common.flow.FlowConstants;
import com.netapp.common.flow.FlowException;
import com.netapp.common.flow.OperationId;
import com.netapp.common.flow.api.OperationResult;
import com.netapp.common.flow.api.OperationService;
import com.netapp.common.flow.api.OperationStartResult;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.StringUtils;

/**
 * Binds abstract methods annotated as {@link WorkflowOperation}s to workflows in a 
 * workflow engine.
 * 
 * Requires CGLIB libraries available on the classpath.
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class WorkflowBindingFactory {

    private static final NALogger logger = NALogger.getLogger();
    private Engine engine;
    private ResultDelegate resultDelegate;
    private List<WorkflowBindingHandler> bindingHandlers;
    
    public void setEngine(Engine engine) {
        this.engine = engine;
    }
    
    public void setResultDelegate(ResultDelegate resultDelegate) {
        this.resultDelegate = resultDelegate;
    }
    /**
     * Construct a new implementation class based upon the input abstract class.
     * Any abstract methods with {@link WorkflowOperation} annotations in the abstract class 
     * will be delegated to the supplied workflow engine.
     * 
     * @param className An Abstract class satisfying the following requirements that 
     * implements the OperationService interface (or an extension thereof)
     * @return Implementation for the supplied abstract class (WorkflowOperation methods filled in).
     */
    public Object bindWorkflows(String className) {
        try {
            Class<?> clazz = Class.forName(className);
            return bindWorkflows(clazz);
        }
        catch (ClassNotFoundException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_SOAP_SERVICE_CLASS_NOT_FOUND, className));
        }
    }

    public Object bindWorkflows(Class<?> clazz) {
        assert this.engine != null;
        WorkflowDiagnostic diag = clazz.getAnnotation(WorkflowDiagnostic.class);
        EngineConfiguration configuration = this.engine.getConfiguration();
        if (diag != null) {
            configuration.addOperation(FlowConstants.DIAGNOSTIC_OPERATION_NAME, diag.value());
        }
        
        for (Method m : clazz.getMethods()) {
            WorkflowOperation op = m.getAnnotation(WorkflowOperation.class);
            if (op == null) {
                continue;
            }
            
            if (!Modifier.isAbstract(m.getModifiers())) {
                throw new FlowException(logger.error(MsgKey.FLOW_SOAP_WORKFLOW_METHOD_NOT_ABSTRACT, m.toString()));
            }
            
            configuration.addOperation(ServiceImplFactory.getWebMethodName(m), op.value());
        }
        
        this.engine.start();
        
        MethodInterceptor serviceMethodInterceptor = new WorkflowMethodInterceptor(this.engine, this.resultDelegate, getBindingHandlers());
        Enhancer enhancer = new Enhancer();
        enhancer.setSuperclass(clazz);
        enhancer.setCallback(serviceMethodInterceptor);
        
        return enhancer.create();
    }
    
    protected List<WorkflowBindingHandler> getBindingHandlers() {
        return this.bindingHandlers;
    }

    public void setBindingHandlers(List<WorkflowBindingHandler> bindingHandlers) {
        this.bindingHandlers = bindingHandlers;
    }

    /**
     * CGLIB method interceptor used by {@link #createServiceImplFromAbstractClass}.
     * Delegates WorkflowOperation methods to workflow engine
     */
    private static class WorkflowMethodInterceptor implements MethodInterceptor {
        private Engine engine;
        private OperationServiceImpl operationService;
        private List<WorkflowBindingHandler> bindingHandlers;

        public WorkflowMethodInterceptor(Engine engine, ResultDelegate resultDelegate, List<WorkflowBindingHandler> bindingHandlers) {
            this.engine = engine;
            this.bindingHandlers = bindingHandlers;
            this.operationService = new OperationServiceImpl(this.engine);
            this.operationService.setResultDelegate(resultDelegate);
        }
        
        public Object intercept(Object obj, Method method, Object[] paramArray, MethodProxy methodProxy) throws Throwable {
            if (!Modifier.isAbstract(method.getModifiers())) {
                return methodProxy.invokeSuper(obj, paramArray);
            }
            
            if (method.getDeclaringClass().equals(OperationService.class)) {
                return method.invoke(this.operationService, paramArray);
            }
            
            // Custom result object
            if (method.getName().endsWith("Result")
                    && OperationResult.class.isAssignableFrom(method.getReturnType())
                    && paramArray.length == 1) {
                return doResult(method, paramArray);  
            }
            
            // Otherwise it's a OperationStart call, so put a request together for the
            // workflow engine.
            OperationStartResult startResult = new OperationStartResult();
            try {
                startResult.setId(doStart(method, paramArray).getId());
            }
            // CHECKSTYLE:OFF
            catch (Throwable t) {
            // CHECKSTYLE:ON
                startResult.setError(t.getLocalizedMessage());
            }
            return startResult;

        }
        
        @SuppressWarnings("unchecked") // Class cast is unchecked
        private OperationResult doResult(Method method, Object[] paramArray) throws Throwable {
            Class<? extends OperationResult> clazz = (Class<OperationResult>)method.getReturnType();
            return this.operationService.operationResult(paramArray[0].toString(), clazz);
        }
        
        private OperationId doStart(Method method, Object[] paramArray) throws Throwable {
            if (!OperationStartResult.class.equals(method.getReturnType())) {
                throw new FlowException(logger.error(MsgKey.FLOW_SOAP_SERVICE_INVALID_METHOD_RETURN, method.toString()));
            }

            /*
             * TODO - remove this coupling to JAXWS annotations
             */
            String operationName = ServiceImplFactory.getWebMethodName(method);
            if (StringUtils.isEmpty(operationName)) { 
                throw new FlowException(logger.error(MsgKey.FLOW_SOAP_SERVICE_NOT_WEB_METHOD, method.toString()));
            }

            Annotation[][] annotations = method.getParameterAnnotations();
            Map<String, Object> params = new HashMap<String, Object>();
            
        param: 
            for (int i = 0; i < annotations.length; i++) {
                for (Annotation a : annotations[i]) {
                    if (a instanceof WorkflowParam) {
                        WorkflowParam wpa = (WorkflowParam)a;
                        
                        params.put(wpa.value(), paramArray[i]);
                        continue param;
                    }

                    throw new FlowException(logger.error(MsgKey.FLOW_SOAP_SERVICE_MISSING_PARAMETER, method.toString(), i, WorkflowParam.class));
                }
            }
            
            // The user can set header annotations on the method itself. We will use those annotations to pull values from the request
            // object and store them into registers.
            for (Annotation annotation : method.getAnnotations()) {
                if (annotation instanceof WorkflowContextParams) {
                    WorkflowContextParams workflowHeaders = (WorkflowContextParams) annotation;
                    
                    for (WorkflowContextParam workflowHeader : workflowHeaders.value()) {
                        for (WorkflowBindingHandler handler : this.bindingHandlers) {
                            if (handler.canHandleType(workflowHeader.type())) {
                                params.put(workflowHeader.register(), handler.getValue(workflowHeader.name()));
                            }
                        }
                    }
                }
            }
        
            return this.engine.operationStart(operationName, params);
        }
    }
}
