/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/DeviceAnalyzer.java#1 $
 * 
 * Copyright (c) 2008 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform;

import java.io.File;

public interface DeviceAnalyzer {

    /**
     * Determine if a file is a recognized raw or block device.
     * The results are undefined if file does not exist in the 
     * active file system or the current process is not running 
     * as root.
     * 
     * @param file file to check
     * @return Device type of the file
     */
    DeviceType determineDeviceType(File file);
    
    /**
     * Extract the partition number from the device path
     * @param file Device path (e.g. /dev/sdb1)
     * @return The partition number (1 is the first partition)
     * @throws UnsupportedOperationException if the operating system does not natively support 
     * partitions (e.g. HP-UX and AIX) 
     */
    int extractPartitionNumber(File file);
    
    /**
     * Extract the partition identifier from a given partition
     * block device. <code>file</code> must have previously been
     * determined to be a {@link DeviceType#PARTITION} with a call
     * to {@link #determineDeviceType(File)}.
     * 
     * The identifier returned is system-specific, and should be
     * treated as an opaque object. It may be used to locate the
     * same partition on a cloned device using {@link #deviceForPartition(File, String)}.
     * Specifically, it is not guaranteed that <code>file.getAbsolutePath() ==
     * extractFullDiskPathElement(file) + extractPartitionElement(file)</code>.
     * @param file
     * @return The opaque partition identifier portion of the supplied partition block device. 
     * If <code>file</code> is not a partition block device, the result is undefined.
     * @see #extractFullDiskPathElement(File)
     * @see #deviceForPartition(File, String)
     */
    String extractPartitionPathElement(File file);
    
    /**
     * Extract the underlying whole disk device from a given
     * partition block device. <code>file</code> must have previously
     * been determined to be a {@link DeviceType#PARTITION} with a
     * call to {@link #determineDeviceType(File)}.
     *
     * @param file Device path (e.g. <code>/dev/sdb1</code>)
     * @return The path of the whole disk block device for the given
     * partition (e.g. <code>/dev/sdb</code>). If <code>file</code> is not a partition block device,
     * the result is undefined.
     */
    String extractFullDiskPathElement(File file);
    
    /**
     * Construct a partition device (/dev/sda1) from the
     * enclosing device and partition identifier.
     * 
     * Various platforms have different rules on constructing partition devices,
     * use this function instead of <code>disk.getAbsolutePath() + partition</code>.
     * 
     * @param disk Whole disk device
     * @param partition Partition identifier previously obtained from {@link #extractPartitionPathElement(File)}
     * @return partition block device.
     * @see #extractFullDiskPathElement(File)
     * @see #extractPartitionPathElement(File)
     */
    File deviceForPartition(File disk, String partition);
    
    /**
     * Construct a partition device (/dev/sda1) from the
     * enclosing device and partition number.
     * 
     * Various platforms have different rules on constructing partition devices,
     * use this function instead of <code>disk.getAbsolutePath() + partition</code>.
     * 
     * @param disk Whole disk device
     * @param partitionNumber Partition identifier previously obtained from {@link #extractPartitionPathElement(File)}
     * @return The complete block device for the specified disk and partition number
     * @see #extractFullDiskPathElement(File)
     * @see #extractPartitionPathElement(File)
     */
    File deviceForPartition(File disk, int partitionNumber);
    
    /**
     * Lookup the block device backing a raw device
     * @param rawDevice
     * @return Raw device mapping for the raw / block device pair
     */
    RawDeviceMapping lookupMappingByRawDevice(File rawDevice);

    /**
     * Lookup the raw device attached to a given block device
     * @param disk
     * @return Raw device mapping for the raw / block device pair
     */
    RawDeviceMapping lookupMappingByDisk(File disk);
}
