/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/linux/DOSPartitionTable.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.linux;


import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.platform.DeviceAnalyzer;
import com.netapp.common.platform.OperatingSystem;
import com.netapp.common.platform.Partition;
import com.netapp.common.platform.PartitionTable;
import com.netapp.common.util.Utils;

/**
 * Read a DOS-style Partition Table.
 * Implementation based on the following
 * <a href="http://ata-atapi.com/hiwtab.htm">partition table reference</a>
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
class DOSPartitionTable extends PartitionTable {

    private static final NALogger logger = NALogger.getLogger(DOSPartitionTable.class);
    
    private static final int PARTITION_TABLE_START = 446;
    private static final int PARTITION_TABLE_SIZE = 66;

    private static final int PARTITION_TABLE_ENTRY_SIZE = 16;
    private static final int NUM_PARTITION_TABLE_ENTRIES = 4;
        
    private static final int SIZE_OFFSET = 12;
    private static final int LBA_OFFSET = 8;
    private static final int TYPE_OFFSET = 4;
    
    private static final int BYTES_PER_SECTOR = 512;

    private byte[] partitionTable;
    
    public DOSPartitionTable(File device, OperatingSystem operatingSystem) throws IOException {
        RandomAccessFile randomAccessFile = new RandomAccessFile(device, "r");
        
        FileChannel fileChannel = randomAccessFile.getChannel();
        
        this.partitionTable = new byte[PARTITION_TABLE_SIZE];
        ByteBuffer blockBytesBuffer = ByteBuffer.wrap(this.partitionTable);

        fileChannel.position(PARTITION_TABLE_START);
        
        fileChannel.read(blockBytesBuffer);
        
        for (int i = 0; i < NUM_PARTITION_TABLE_ENTRIES; i++) {
            int start = i * PARTITION_TABLE_ENTRY_SIZE;
            /*
             * The first byte read is the low-order byte of the long
             */
            long lba = Utils.unsigned32LE(this.partitionTable, start + LBA_OFFSET);

            long size = Utils.unsigned32LE(this.partitionTable, start + SIZE_OFFSET);
            DeviceAnalyzer analyzer = operatingSystem.createDeviceAnalyzer();
            
            if (lba != 0 && size != 0) {
                File devicePartition = analyzer.deviceForPartition(device, Integer.toString(i + 1));
                this.addPartition(new Partition(devicePartition, lba * BYTES_PER_SECTOR, size * BYTES_PER_SECTOR));
            }
        }
        
        ByteArrayOutputStream stringOutputStream = new ByteArrayOutputStream();
        this.print(new PrintStream(stringOutputStream));
        
        logger.debug(MsgKey.PLATFORM_PARTITION_TABLE, device, stringOutputStream);
    }

    public void print(PrintStream out) {
        out.println(String.format("#, active, type, start, end, size"));
        for (int i = 0; i < NUM_PARTITION_TABLE_ENTRIES; i++) {
            int start = i * PARTITION_TABLE_ENTRY_SIZE;
            boolean active = (this.partitionTable[start] & ((byte)0x80)) != 0;
            short type = Utils.unsigned8(this.partitionTable[start + TYPE_OFFSET]);
            
            long lba = Utils.unsigned32LE(this.partitionTable, start + LBA_OFFSET);
            long size = Utils.unsigned32LE(this.partitionTable, start + SIZE_OFFSET);
            
            if (type == 0 || (lba == 0 && size == 0)) {
                out.println(i + " NOT USED");
            }
            else {
                out.println(String.format("%d, %s, %x, %d, %d", i, (active ? "A" : "-"), type, lba, lba + size - 1, size));
            }
        }
    }
}
