/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/remote/RemoteObjectFactoryRegistryImpl.java#1 $
 * 
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.remote;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.rmi.AccessException;
import java.rmi.AlreadyBoundException;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.netapp.common.logging.NALogger;

/**
 * A Registry for RemoteObjectFactory's.  By default a Registry does not allow an object to be
 * bound from a remote host.  This class will allow it if the remote object is a RemoteObjectFactory.
 * 
 * @author claird
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class RemoteObjectFactoryRegistryImpl  extends UnicastRemoteObject implements RemoteObjectFactoryRegistry {

    private static final NALogger logger = NALogger.getLogger();
    
    private int registryPort;
    
    private Map<String, Remote> bindings;
    
    /**
     * Specifies whether or not to create or reuse the registry on the given port 
     */
    private Registry bootstrapRegistry;
    
    static final String REGISTRY_BIND_NAME = "remote.object.factory.registry";
    
    public RemoteObjectFactoryRegistryImpl(int registryPort) throws RemoteException {
        try {
            bindings = new ConcurrentHashMap<String, Remote>();
            logger.internal("RemoteObjectFactoryRegistryImpl: initializing on port %s", registryPort);
            this.bootstrapRegistry = LocateRegistry.createRegistry(registryPort);
            this.registryPort = registryPort;
            bootstrapRegistry.bind(REGISTRY_BIND_NAME, this);
        }
        catch (AlreadyBoundException e) {
            throw new RuntimeException(e);
        }
    }

    public static RemoteObjectFactoryRegistry getRegistry(String registryHost, int registryPort) {
        logger.internal("RemoteObjectFactoryRegistryImpl: getting registry at //%s:%s", registryHost, registryPort);
        
        try {
            Registry bootstrapRegistry = LocateRegistry.getRegistry(registryHost, registryPort);
            RemoteObjectFactoryRegistry instance = (RemoteObjectFactoryRegistry) bootstrapRegistry.lookup(REGISTRY_BIND_NAME);
            return instance;
        }
        catch (RemoteException e) {
            throw new RuntimeException(e);
        }
        catch (NotBoundException e) {
            throw new RuntimeException(e);
        }
    }
    
    public static RemoteObjectFactoryRegistry getRegistry(int registryPort) {
        try {
            return getRegistry(InetAddress.getLocalHost().getHostAddress(), registryPort);
        }
        catch (UnknownHostException e) {
            throw new RuntimeException(e);
        }
    }
    
    /**
     * @see com.netapp.common.remote.RemoteObjectFactoryRegistry#destroy()
     */
    public void destroy() {
        try {
            logger.internal("RemoteObjectFactoryRegistryImpl: shutting down on port %s", this.registryPort);
            UnicastRemoteObject.unexportObject(this, true);
            UnicastRemoteObject.unexportObject(this.bootstrapRegistry, true);
        }
        catch (RemoteException e) {
            logger.internal("Ignoring exception while destroying RemoteObjectFactoryRegistryImpl.", e);
        }
    }

    /**
     * @see java.rmi.registry.Registry#bind(java.lang.String, java.rmi.Remote)
     */
    public void bind(String name, Remote obj) throws RemoteException, AlreadyBoundException {
        if (this.bindings.containsKey(name)) {
            throw new AlreadyBoundException(name);
        }
        
        checkType(obj);
        this.bindings.put(name, obj);
    }

    /**
     * @see java.rmi.registry.Registry#list()
     */
    public String[] list() throws RemoteException {
        return this.bindings.keySet().toArray(new String[0]);
    }

    /**
     * @see java.rmi.registry.Registry#lookup(java.lang.String)
     */
    public Remote lookup(String name) throws RemoteException, NotBoundException {
        Remote obj = this.bindings.get(name);
        if (obj == null) {
            throw new NotBoundException(name);
        }
        
        return obj;
    }

    /**
     * @see java.rmi.registry.Registry#rebind(java.lang.String, java.rmi.Remote)
     */
    public void rebind(String name, Remote obj) throws RemoteException {
        checkType(obj);
        this.bindings.put(name, obj);
    }

    /**
     * @see java.rmi.registry.Registry#unbind(java.lang.String)
     */
    public void unbind(String name) throws RemoteException, NotBoundException {
        Remote obj = this.bindings.remove(name);
        if (obj == null) {
            throw new NotBoundException(name);
        }
    }

    private void checkType(Remote obj) throws AccessException {
        if (!(obj instanceof RemoteObjectFactory)) {
            throw new AccessException("Only RemoteObjectFactory's allowed in RemoteObjectFactoryRegistryImpl");
        }
    }

    
    @Override
    protected void finalize() throws Throwable {
        this.bindings.clear();
        super.finalize();
    }
}
