/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/AdapterFactory.java#1 $
 * 
 * Copyright (c) 2008 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import com.netapp.common.logging.MessageBundle;
import com.netapp.common.MsgKey;

import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.MethodInterceptor;
import net.sf.cglib.proxy.MethodProxy;

/**
 * Factory for creating adapter classes at runtime (and objects thereof).
 * An "adapter" is a class that implements a given interface
 * by forwarding method calls to an internal "adaptee" object.
 * 
 * <ul>
 * <li>Method signatures of any abstract methods in the adapter class 
 * that match the method signatures of the adaptee will be forwarded to the adaptee.
 * <li>Method signatures of any abstract methods in the adapter class
 * that do not match the method signatures of the adaptee will yield 
 * {@link NoSuchMethodException} when invoked. 
 * <li>Concrete methods defined in the adapter class (if any) will remain unaffected.
 * </ul>
 * 
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class AdapterFactory {
    
    private static final AdapterMethodInterceptor ADAPTER_METHOD_INTERCEPTOR = new AdapterMethodInterceptor();

    private AdapterFactory() {
    }
    
    public interface Adapter<T> {

        /**
         * Get the adaptee that this Adapter is adapting.
         */
        T getAdaptee();
        
        /**
         * Set the adaptee that this Adapter is adapting.
         * Called immediately after construction. 
         */
        void setAdaptee(T adaptee);
        
        /**
         * Handle the error case where the adaptee is null.  If this is not
         * defined by an abstract class, an IllegalStateException is thrown
         * in this case.
         */
        void handleNullAdaptee();

        /**
         * Hook allowing the Adapter to alter the pre or post behavior of
         * any of the adapted methods.
         * @param adapteeInvoker adapter should call invoke on this object to perform the actual method invocation
         * @param adaptee the object being adapted
         * @param adapteeMethod the method being adapted
         * @param args the args to use for the adapteeMethod
         * 
         * @return The adapted result of the invocation (if any)
         */
        Object adapt(AdapteeInvoker adapteeInvoker, Object adaptee, Method adapteeMethod, Object[] args) throws Throwable;
    }
    
    /**
     * An object capable of invoking an adaptee method.
     * 
     * @author sirianni
     * @version $Revision: #1 $ $Date: 2010/08/20 $
     */
    public interface AdapteeInvoker {
        
        Object invoke(Object adaptee, Method adapteeMethod, Object[] args) throws Throwable;
        
    }
    
    @SuppressWarnings("unchecked")
    public static <T extends Adapter> T buildAdapter(Class<T> clazz, Object adaptee) {
        Enhancer enhancer = new Enhancer();
        enhancer.setSuperclass(clazz);
        
        // Since the AdapterMethodInterceptor maintains no state, we can just use a singleton
        enhancer.setCallback(ADAPTER_METHOD_INTERCEPTOR);
        
        T adapter = (T) enhancer.create();
        adapter.setAdaptee(adaptee);
        return adapter;
    }

    /**
     * Intercepts the abstract methods of a given Adapter and forwards to an adaptee object
     * @author sirianni
     * @version $Revision: #1 $ $Date: 2010/08/20 $
     */
    private static class AdapterMethodInterceptor implements MethodInterceptor, AdapteeInvoker {

        public Object intercept(Object object, Method adapterMethod, Object[] args, MethodProxy methodProxy) throws Throwable {
            if (!Modifier.isAbstract(adapterMethod.getModifiers())) {
                return methodProxy.invokeSuper(object, args);
            }
            else if (adapterMethod.getName().equals("adapt")) {
                /*
                 * Do a fake no-op adapt() if the class being extended forgot to define adapt()
                 */
                AdapteeInvoker adapteeInvoker = (AdapteeInvoker) args[0];
                Object adaptee = args[1];
                Method adapteeMethod = (Method) args[2];
                Object[] invocationArgs = (Object[]) args[3];
                
                return adapteeInvoker.invoke(adaptee, adapteeMethod, invocationArgs);
            }
            else if (adapterMethod.getName().equals("handleNullAdaptee")) {
                /*
                 * If the class being extended did not define
                 * handleAdaptee(), throw an IllegalStateException since the
                 * adaptee is null. 
                 */
                throw new IllegalStateException(MessageBundle.localizeMessage(MsgKey.UTIL_NULL_ADAPTEE, object.getClass().getName()));
            }
            
            Adapter<?> adapter = (Adapter<?>) object;
            if (adapter.getAdaptee() == null) {
                adapter.handleNullAdaptee();
            }
            Class<?> adapteeClass = adapter.getAdaptee().getClass();
            
            /*
             * Find the adaptee method that matches the signature of the method invoked on the adapter
             */
            Method adapteeMethod = adapteeClass.getMethod(adapterMethod.getName(), adapterMethod.getParameterTypes());
            
            /*
             * Even if the method is public, method.invoke() will throw an IllegalAccessException
             * if the class itself is only package visible.  This is stupid.
             * 
             * http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6342411
             * 
             * Call setAccessible(true) to work around this.
             */
            adapteeMethod.setAccessible(true);
            
            /*
             * Make sure the return types match
             */
            if (!adapterMethod.getReturnType().isAssignableFrom(adapteeMethod.getReturnType())) {
                throw new NoSuchMethodException(adapterMethod.toString());
            }
            
            /*
             * Technically, the adaptee method isn't a subtype of the adapter method if the Exceptions
             * thrown by the adaptee method aren't a subset of the Exceptions thrown by the adapter method.
             * 
             * But let's ignore that.  
             * I assume CGLIB softens the Throwable into a RuntimeException if this happens??
             */
            
            return adapter.adapt(this, adapter.getAdaptee(), adapteeMethod, args);
        }

        public Object invoke(Object adaptee, Method adapteeMethod, Object[] args) throws Throwable {
            try {
                return adapteeMethod.invoke(adaptee, args);
            }
            catch (InvocationTargetException e) {
                throw e.getCause();
            }
        }

    }

}

