/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/bootstrap/com/netapp/common/util/StringUtils.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.text.NumberFormat;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.builder.ReflectionToStringBuilder;
import org.apache.commons.lang.builder.ToStringStyle;


/**
 * Simple string manipulation routines.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class StringUtils {
    
    private StringUtils() {
        throw new RuntimeException("StringUtils may not be instantiated");
    }
    
    /**
     * Test if the supplied string is null or empty (i.e "").
     * 
     * @param string String to test
     * @return True if the string is null or empty
     */
    public static boolean isEmpty(String string) {
        return string == null || string.equals("");
    }
    
    /**
     * Test if the supplied string is null or blank (i.e contains only whitespace).
     * 
     * @param string String to test
     * @return True if the string is null or blank
     */
    public static boolean isBlank(String string) {
        return string == null || string.trim().equals("");
    }
    
    /**
     * Convert the first character to upper case
     * @param string input string
     * @return string with initial upper case
     */
    public static String initialUpperCase(String string) {
        StringBuffer b = new StringBuffer();;
        if (string.length() > 0) {
            b.append(string.substring(0, 1).toUpperCase());
        }
        if (string.length() > 1) {
            b.append(string.substring(1));
        }
        return b.toString();
    }
    
    /**
     * Convert the first character of a string to lower case
     * 
     * @param string input string
     * @return string with initial lower case
     */
    public static String initialLowerCase(String string) {
        StringBuffer b = new StringBuffer();
        if (string.length() > 0) {
            b.append(string.substring(0, 1).toLowerCase());
        }
        if (string.length() > 1) {
            b.append(string.substring(1));
        }
        return b.toString();
    }
    
    public static int countInitialUpperCaseChars(String string) {
        int i;
        for (i = 0; i < string.length(); i++) {
            char c = string.charAt(i);
            
            if (!Character.isUpperCase(c)) {
                break;
            }
        }

        return i;
    }
    
    /**
     * Given a list of strings, return the first {@link #isEmpty(String) non-empty} string.
     * @param strings List of strings to check
     * @return The first entry in strings that is not empty
     */
    public static String coalesce(String... strings) {
        String output = null;
        for (String v : strings) {
            if (!isEmpty(v)) {
                output = v;
                break;
            }
        }
        return output;
    }
    
    /**
     * Given a supplied regex pattern and a CharSequence through which to search,
     * this will return a String[] of the matching groups.  If none are found,
     * it will return null.<br><br>
     * Example:<br>
     * <code>
     * String[] matches = StringUtils.findRegExMatches("abc (.+) ghi", "abc def ghi");
     * if (matches != null) {
     *   String myMatch = matches[0];
     * }
     * </code>
     * @param patternStr The regex pattern to search for.
     * @param input The text through which to search for the pattern.
     * @return All of the matching groups; else, null.
     */
    public static String[] findRegExMatches(String patternStr, CharSequence input) {
        Pattern pattern = Pattern.compile(patternStr);
        Matcher matcher = pattern.matcher(input);
        if (matcher.find()) {
            int groupCount = matcher.groupCount();
            String[] result = new String[groupCount];
            for (int i = 0; i < groupCount; i++) {
                result[i] = matcher.group(i + 1);
            }
            return result;
        }
        return null;
    }
    
    public static boolean containsAny(String s, String... substrings) {
        for (String substring : substrings) {
            if (s.contains(substring)) {
                return true;
            }
        }
        return false;
    }

    public static boolean isUpperCase(String s) {
        return s.equals(s.toUpperCase());
    }
    
    
    /**
     * Tests if a string contains any characters in the Collection of chars
     * @param s The string to check
     * @param chars The collection of characters we are checking for
     * @return true if the string contains any of the characters in the collection
     */
    public static boolean stringContainsChar(String s, Collection<Character> chars) {
        for (Character character : chars) {
            if (s.indexOf(character) >= 0) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * Creates a String representation of n containing the number padded with "0"
     * to contain the number of digits specified.
     * @param n The number to create a String representation of
     * @param digits The number of digits the String created should contain
     * @return String representation of n
     */
    public static String paddedToString(int n, int digits) {
        String s = "" + n;
        if (s.length() < digits) { // pad on left with zeros
            s = "000000000000000000000000000000".substring(0, digits - s.length()) + s;
        }
        return s;
    }
    
    /**
     * Creates a String padded on the right with space characters.
     * @param s The String to be padded
     * @param size The size the String returned should b
     * @return String padded to the size specified
     */
    public static String paddedToString(String s, int size) {
        if (s.length() < size) { // pad on right with spaces
            s = s + "                        ".substring(0, size - s.length());
        }
        return s;
    }
    
    /**
     * Creates a String padded on the left with space characters.
     * @param s The String to be padded
     * @param size The size the String returned should b
     * @return String padded to the size specified
     */
    public static String leftPaddedToString(String s, int size) {
        if (s.length() < size) { // pad on left with spaces
            s = "                        ".substring(0, size - s.length()) + s;
        }
        return s;
    }
    
    /**
     * Creates a String representation of the object using reflection.
     * 
     * @param o The object
     * @return String representation of the object specified
     */
    public static String reflectiveToString(Object o) {
        if (o == null) {
            return "null";
        }
        
        if (o instanceof String) {
            return "String[value=" + (String) o + "]";
        }
        
        
        ToStringStyle toStringStyle = new ToStringStyle() {
            {
                this.setUseShortClassName(true);
                this.setUseIdentityHashCode(false);
            }
            @Override
            public void append(StringBuffer buffer, String fieldName, Object fieldValue, Boolean fullDetail) {
                if (fieldValue instanceof String || fieldValue instanceof Integer 
                        || fieldValue instanceof Boolean || fieldValue instanceof Long) {
                    super.append(buffer, fieldName, fieldValue, fullDetail);
                }
                else {
                    super.append(buffer, fieldName, 
                        StringUtils.reflectionToString(fieldValue, this), fullDetail);
                }
            }
        };
        
        return reflectionToString(o, toStringStyle);
    }

    private static String reflectionToString(Object object, ToStringStyle toStringStyle) {
        ReflectionToStringBuilder reflectionToStringBuilder = new ReflectionToStringBuilder(object,
            toStringStyle,
            null,         // buffer
            Object.class, // up to class
            false,        // transients?
            false);       // statics?

        reflectionToStringBuilder.setExcludeFieldNames(new String[] {"password"});
        
        return reflectionToStringBuilder.toString();
    }
    
    /**
     * returns a Set containing the String representation of each object in stringArray
     * @param array of objects to get String representations of
     * @return Set
     */
    public static Set<String> toStringSet(Object[] stringArray) {
        if (stringArray == null) {
            return null;
        }
        else {
            Set<String> stringSet = new HashSet<String>();
            for (int i = 0; i < stringArray.length; i++) {
                stringSet.add(stringArray[i].toString());
            }
            return stringSet;
        }
    }
    
    /**
     * Given a Collection and a String delimimiter, creates a String representation.
     * @param c
     * @param delimiter
     * @return String representation of Collection
     */
    public static String collectionToString(Collection<? extends Object> c, String delimiter) {
        StringBuilder stringBuilder = new StringBuilder();
        for (Iterator i = c.iterator(); i.hasNext();) {
            Object o = i.next();
            stringBuilder.append(o.toString());
            if (i.hasNext()) {
                stringBuilder.append(delimiter);
            }
        }
        return stringBuilder.toString();
    }
    
    private static final NumberFormat NUMBER_FORMAT = NumberFormat.getInstance();

    static {
        NUMBER_FORMAT.setGroupingUsed(false);
    }

    /**
     * Convert a number to a string (without commas)
     * @param number A number (e.g 1234)
     * @return A string representing that number without commas (e.g. 1234 not 1,234)
     */
    public static String numberToStringNoCommas(Object number) {
        return NUMBER_FORMAT.format(number);
    }
    
    /**
     * Determines if a String ends with a suffix (ignoring case).
     * @param s String to test
     * @param suffix Suffix to test
     * @return true if String ends with suffix (ignoring case).  false otherwise
     */
    public static boolean endsWithIgnoreCase(String s, String suffix) {
        if (s.length() < suffix.length()) {
            return false;
        }
        
        String suffixToCheck = s.substring(s.length() - suffix.length(), s.length());
        return suffixToCheck.equalsIgnoreCase(suffix);
    }
    
    /**
     * Returns the number of times c occurs in String s
     * @return the number of times c occurs in String s
     */
    public static int countCharactersInString(String s, char c) {
        if (s == null) {
            return 0;
        }
        int counter = 0;
        for (int i = 0; i < s.length(); i++) {
            if (s.charAt(i) == c) {
                counter++;
            }
        }
        return counter;
    }
    

    /**
     * Splits string using the space character as a delimeter, then recombines sub-strings after capitalizing 
     * first char in each sub-string.<br>
     * Example:<br>
     * "quick brown fox" becomes "Quick Brown Fox"
     */
    public static String getCamelCase(String value) {

        String retVal = value;
        if (!StringUtils.isEmpty(value)) {

            value = value.toLowerCase();
            String[] split = value.split(" ");

            StringBuilder master = new StringBuilder();
            for (String string : split) {

                StringBuilder sb = new StringBuilder(string);
                String substring = sb.substring(0, 1).toUpperCase();
                sb.deleteCharAt(0).insert(0, substring);

                master.append(sb.toString()).append(" ");
            }

            retVal = master.toString().trim();
        }

        return retVal;
    }
    
    public static String nullSafe(String s) {
        return s == null ? "" : s;
    }
    
    public static String truncate(String input, int maxLength) {
        if (maxLength < input.length()) {
            input = input.substring(0, maxLength);
        }
        
        return input;
    }
}
