/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/ZAPIExecutorFactory.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi;

import static com.netapp.common.util.ConfigProperty.booleanProperty;
import java.util.EnumMap;
import java.util.Map;

import com.netapp.common.logging.NALogger;
import com.netapp.common.util.Cache;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.ConfigProperty;
import com.netapp.common.util.StringUtils;

/**
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class ZAPIExecutorFactory {
    @SuppressWarnings("unused")
    private static final NALogger logger = NALogger.getLogger(ZAPIExecutorFactory.class);
    public static final String[] DEFAULT_CONFIG_FILES = { "zapi-executor.properties", "zapi-executor.override" };
    public static final String PROPERTY_PREFIX = "zapi.url";
    public static final ConfigProperty<Boolean> JMX_ENABLE_PROPERTY = booleanProperty("zapi.jmx.enable", true);
    public static final String JMX_PREFIX = "com.netapp.common.zapi:";
    
    public enum ZAPITargetType {
        STORAGE("storage"),
        PROTECTION_MANAGER("protectionManager");
        
        private final String configName;
        
        private ZAPITargetType(String configName) {
            this.configName = configName;
        }
        
        protected String getConfigName() {
            return this.configName;
        }
    }

    private final Map<ZAPITargetType, ZAPIHostInfo> defaultHostInfos;
    private Map<ZAPITargetType, Cache<String, ZAPIExecutor>> executorCache;
    private Map<ZAPITargetType, Cache<String, ZAPIRunner>> runnerCache;
    
    private final ConfigBundle configBundle;
    
    private boolean enableJMX;
    
    public ZAPIExecutorFactory() {
        this(new ConfigBundle(DEFAULT_CONFIG_FILES));
    }
    
    public ZAPIExecutorFactory(ConfigBundle configBundle) {
        this.configBundle = configBundle;
        this.enableJMX = JMX_ENABLE_PROPERTY.lookup(configBundle);
        
        this.defaultHostInfos = new EnumMap<ZAPITargetType, ZAPIHostInfo>(ZAPITargetType.class);
        this.executorCache = new EnumMap<ZAPITargetType, Cache<String, ZAPIExecutor>>(ZAPITargetType.class);
        this.runnerCache = new EnumMap<ZAPITargetType, Cache<String, ZAPIRunner>>(ZAPITargetType.class);
        
        for (ZAPITargetType type : ZAPITargetType.values()) {
            ConfigProperty<ZAPIHostInfo> p = new ConfigProperty<ZAPIHostInfo>(ZAPIHostInfo.class, PROPERTY_PREFIX + "." + type.configName);        
            this.defaultHostInfos.put(type, p.lookup(this.configBundle));
        }
        
    }
    
    public void setEnableJMX(boolean useJMX) {
        this.enableJMX = useJMX;
    }
    
    public ZAPIExecutor getZAPIExecutor(String host, ZAPITargetType type) throws InvalidTargetException {
        ZAPIExecutor executor = null;
        
        Cache<String, ZAPIExecutor> cache = null;
        synchronized (this.executorCache) {
            cache = this.executorCache.get(type);
            if (cache == null) {
                cache = new Cache<String, ZAPIExecutor>(this.enableJMX ? JMX_PREFIX + "type=Executor,name=" + type.configName : null);
            }
            this.executorCache.put(type, cache);
        }
        
        synchronized (cache) {
            executor = cache.get(host);
            if (executor != null) {
                return executor;
            }

            if (type.equals(ZAPITargetType.STORAGE)) {
                executor = getStorageExecutor(host);
            }
            else {
                executor = getDfmExecutor(host);
            }
            cache.put(host, executor);
        }
        return executor;
    }

    /**
     * Clears all (invalid) runners from the cache
     * @param type
     */
    public void clearRunnerCache(ZAPITargetType type) {
        try {
            this.runnerCache.get(type).clear();
        } 
        catch (NullPointerException e) {
            // Nothing to do 
        }
    }
    /**
     * Clears an (invalid) runner from the cache
     * @param host
     * @param type
     */
    public void clearRunnerCache(String host, ZAPITargetType type) {
        try {
            this.runnerCache.get(type).remove(host);
        } 
        catch (NullPointerException e) {
            // Nothing to do 
        }
    }
    
    /**
     * @param host Hostname
     * @return Initialized DFM executor for host
     * @throws InvalidTargetException
     */
    protected ZAPIExecutor getDfmExecutor(String host) throws InvalidTargetException {
        ZAPIHostInfo hostInfo = getHostInfo(host, ZAPITargetType.PROTECTION_MANAGER);

        if (null != hostInfo) {
            return new DfmZAPIExecutor(
                StringUtils.isEmpty(hostInfo.getHostname()) ? host : hostInfo.getHostname(),
                hostInfo.getPort(), 
                hostInfo.getUsername(), 
                hostInfo.getPassword());
        }
        
        return new DfmZAPIExecutor(host);
    }

    /**
     * @param host hostname of storage appliance
     * @return ZAPIExecutor for appliance host
     * @throws InvalidTargetException
     */
    protected ZAPIExecutor getStorageExecutor(String host) throws InvalidTargetException {
        ZAPIHostInfo hostInfo = getHostInfo(host, ZAPITargetType.STORAGE);
        
        if (null != hostInfo) {
            if (hostInfo.isProxy()) {
                // proxy through the property file host to the supplied hostname
                ProxyZAPIExecutor executor = new ProxyZAPIExecutor(host, 
                    hostInfo.getHostname(), 
                    hostInfo.getPort(), 
                    hostInfo.getUsername(), 
                    hostInfo.getPassword());
                return executor;
            }
            else {
                return new DirectZAPIExecutor(
                    StringUtils.isEmpty(hostInfo.getHostname()) ? host : hostInfo.getHostname(), 
                    hostInfo.getPort(), 
                    hostInfo.getUsername(), 
                    hostInfo.getPassword());
            }
        }
        
        return new DirectZAPIExecutor(host);
    }
    
    public ZAPIRunner getZAPIRunner(String host, ZAPITargetType type) throws InvalidTargetException {
        ZAPIRunner runner = null;
        Cache<String, ZAPIRunner> cache = null;
        synchronized (this.runnerCache) {
            cache = this.runnerCache.get(type);
            if (cache == null) {
                cache = new Cache<String, ZAPIRunner>(this.enableJMX ? JMX_PREFIX + "type=Runner,name=" + type.configName : null);
                this.runnerCache.put(type, cache);
            }
        }

        synchronized (cache) {
            runner = cache.get(host);
            if (runner == null) {
                runner = new ZAPIRunner(getZAPIExecutor(host, type));
                cache.put(host, runner);
            }
        }
        return runner;
    }
    
    /**
     * Retrieve the ZAPIRunner for the default configured host. If no default is configured, NULL will be returned.
     * 
     * @param type
     * @return ZAPIRunner for the default host
     * @throws InvalidTargetException
     */
    public ZAPIRunner getDefaultZAPIRunner(ZAPITargetType type) throws InvalidTargetException {
        // passing a empty string as the host should never result in a hit which will 
        // force the default host (if it exists) to be returned. If the default does not
        // exist, a null is returned.
        return getZAPIRunner(null, type);
    }
    
    protected ZAPIHostInfo getHostInfo(String host, ZAPITargetType type) {
        ZAPIHostInfo value = defaultHostInfos.get(type);
        
        if (host != null) {
            ConfigProperty<ZAPIHostInfo> prop = new ConfigProperty<ZAPIHostInfo>(ZAPIHostInfo.class, PROPERTY_PREFIX + "." + type.configName + "." + host, value);
            value = prop.lookup(this.configBundle);
        }
        return value;
    }
}
