/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/parser/Api.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi.parser;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Queue;
import java.util.Set;
import java.util.TreeSet;

import org.apache.commons.lang.builder.CompareToBuilder;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.zapi.annotation.StatusEnum;
import com.netapp.common.zapi.generator.GeneratorFactory;

/**
 * Parse and represent a zapi api element and any subtags.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class Api extends Type {
    private static final NALogger logger = NALogger.getLogger(Api.class);
    
    // The docs are silent on the default here. I'm assuming if it wasn't
    // explicitly set to public to hide it
    /**
     * The external status of this api
     */
    StatusEnum status = StatusEnum.UNSUPPORTED;
    
    /**
     * Set of families this api is valid against. Currently
     * only used for documentation.
     * @see com.netapp.common.zapi.annotation.Family
     */
    Set<String> family = new TreeSet<String>();
    
    /**
     * Set of error values this api may return.
     * Currently only used for documentation.
     * @see com.netapp.common.zapi.annotation.Errno
     */
    Set<String> errno = new TreeSet<String>();
    
    /**
     * Construct a new API in the specified category.
     * @param category Category this zapi is defined in
     * @param name Zapi name of this zapi call
     */
    public Api(ZapidocParser parser, String name) {
        super(parser, name);
    }
    
    /**
     * Retrieve the field specified from this api's parameters
     * @param fieldName name of field to retrieve
     * @return Field with the specified zapi name
     */
    protected ApiField getField(String fieldName, boolean input, boolean output) {
        ApiField field = null;
        for (Field f : fields) {
            if (fieldName.equals(f.getName())
                    && ((ApiField) f).input == input
                    && ((ApiField) f).output == output) {
                field = (ApiField) f;
                break;
            }
        }
        return field;      
    }
    
    public Collection<Field> getInputFields() {
        List<Field> results = new ArrayList<Field>();
        for (Field f : fields) {
            if (((ApiField)f).input) {
                results.add(f);
            }
        }
        return results;
    }
    
    public Collection<Field> getOutputFields() {
        List<Field> results = new ArrayList<Field>();
        for (Field f : fields) {
            if (((ApiField)f).output) {
                results.add(f);
            }
        }
        return results;
    }
    
    /**
     * Add the supplied field to this api's list.
     * <ul>
     *   <li> Check for an existing field of the same name. If one exists of the same type,
     *        or it's bit fields with the new one.
     *   <li> If the types differ and one is input while one is output, munge the java names
     *   <li> If unable to rename the fields, log an error
     * </ul>
     * @param field Parameter to add to this api
     */
    private void addField(ApiField field) {
        ApiField existing = getField(field.getName(), field.input, field.output);
        
        if (existing != null && existing.getType().equals(field.getType())) {
            logger.info(MsgKey.ZAPI_GENERATOR_API_FIELD_DEFINED_TWICE, this.getName(), field.getName());
            existing.input |= field.input;
            existing.output |= field.output;
            existing.optional |= field.optional;
        }
        else {
            this.fields.add(field);
        }
        
    }
    
    public boolean hasMultipleFieldsNamed(String name) {
        
        boolean foundOne = false;
        for (Field f : this.fields) {
            if (f.getName().equals(name)) {
                if (foundOne) {
                    return true;
                }
                foundOne = true;
            }
        }
        return false;
        
    }
    
    
    public StatusEnum getStatus() {
        return this.status;
    }
    
    /**
     * Set the public status of this api.
     * @param status public status of this api. Must be one of the types
     *        defined in {@link com.netapp.common.zapi.annotation.StatusEnum}
     */
    protected void setStatus(String status) {
        try {
            this.status = StatusEnum.parse(status);
        }
        catch (IllegalArgumentException e) {
            this.parser.appendError(MsgKey.ZAPI_GENERATOR_API_STATUS_TYPE_UNKNOWN, status, this);
        }
    }
    
    /**
     * Populate elements from the supplied queue of tags, stopping at the first
     * unknown tag. 
     * @param q Queue of tags to parse to populate this api. Any parsed tags are removed.
     */
    @Override
    public void populate(Queue<Tag> q) {
        logger.internal("Processing Api " + name);
        Tag tag = q.peek();
        while (tag != null) {
            ApiField field = null;
            Typedef typedef = null;
            if (tag.is(TagEnum.DESC)) {
                this.setDesc(tag.getValue());
            }
            else if (tag.is(TagEnum.NAMESPACE)) {
                this.setNamespace(tag.getValue());
            }
            else if (tag.is(TagEnum.STATUS)) {
                this.setStatus(tag.getValue());
            }
            else if (tag.is(TagEnum.FAMILY)) {
                for (String f : tag.getValue().split(",")) {
                    this.family.add(f.trim());
                }
            }
            else if (tag.is(TagEnum.ERRNO)) {
                this.errno.add(tag.getFirstWord());
            }
            else if (tag.is(TagEnum.INPUT)) {
                field = new ApiField(tag.getFirstWord());
                field.input = true;
            }
            else if (tag.is(TagEnum.OUTPUT)) {
                field = new ApiField(tag.getFirstWord());
                field.output = true;
            }
            else if (tag.is(TagEnum.TYPEDEF)) {
                typedef = new Typedef(this.parser, tag.getFirstWord());

            }
            else if (tag.is(TagEnum.ATTRIBUTE)
                    || tag.is(TagEnum.DISCUSSION)
                    || tag.is(TagEnum.DOMAIN)
                    || tag.is(TagEnum.JAVA)
                    || tag.is(TagEnum.LICENSE)
                    || tag.is(TagEnum.STREAMING)
                    || tag.is(TagEnum.TEST) 
                    || tag.is(TagEnum.VERSION)
                    || tag.is(TagEnum.VFILER)) {
                // Not recording this
            }
            else {
                logger.internal("Stopping Api on " + tag.getType());
                break;
            }
            // Remove the one we were looking at
            q.poll();
            
            // This is kind of hacky, but we want to process this after we
            // consume the tag that created it.
            if (field != null) {
                field.populate(q);
                
                addField(field);
            }
            if (typedef != null) {
                typedef.populate(q);
            }
            // Peek at the next one
            tag = q.peek();
        }
        if (this.namespace == null) {
            logger.info(MsgKey.ZAPI_GENERATOR_API_DEFAULT_NAMESPACE, this);
            this.setNamespace("filer");
        }
    }

    public Set<String> getErrors() {
        return this.errno;
    }
    
    public Set<String> getFamily() {
        return this.family;
    }
    
    @Override
    protected void addWordReplacements(GeneratorFactory<?, ?, ?, ?> generatorFactory) {
        super.addWordReplacements(generatorFactory);
        if (this.name.endsWith("-start")
                || this.name.endsWith("-next")
                || this.name.endsWith("-end")) {
            
            /*
             * Sometimes the iterator apis are referred to as a group with -* at the end
             */
            generatorFactory.getDescriptionGenerator().addWordReplacement(
                this.name.replaceAll("(-start|-next|-end)", "-*"),
                generatorFactory.getTypeGenerator(this).getNameGenerated().replaceAll("(Start|Next|End)", "*"));
            /*
             * Sometimes the iterator apis are referred to as a group with nothing at the end
             */
            generatorFactory.getDescriptionGenerator().addWordReplacement(
                this.name.replaceAll("(-start|-next|-end)", ""), 
                generatorFactory.getTypeGenerator(this).getNameGenerated());
        }
    }
    
    /**
     * {@link Field} extended to support Input and Output modifiers.
     * @author sunday
     * @version $Revision: #1 $ $Date: 2010/08/20 $
     */
    public class ApiField extends Field {
        boolean input;
        boolean output;
        
        /**
         * Create a new field with the specified name in the enclosing Api object
         * @param name zapi name of the new field to create
         */
        public ApiField(String name) {
            super(Api.this, name);
        }
        
        public boolean isInput() {
            return this.input;
        }
        
        public boolean isOutput() {
            return this.output;
        }
        
//        /**
//         * True if this is an input field. If it's an output only
//         * there's no point making a withXXX setter.
//         * @see com.netapp.common.zapi.parser.Field#isChainable()
//         */
//        @Override
//        public boolean isChainable() {
//            return this.input;
//        }
        
        @Override
        public int compareTo(Field obj) {
            if (obj instanceof ApiField) {
                ApiField apiField = (ApiField) obj;
                
                return new CompareToBuilder()
                    .appendSuper(super.compareTo(obj))
                    .append(this.input, apiField.input)
                    .append(this.output, apiField.output)
                    .toComparison();
            }
            else {
                return super.compareTo(obj);
            }
            
            
        }
    }

}
